# test_brick -- Test nxt.brick module
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
from unittest.mock import Mock, call, patch

import pytest

import nxt.brick
import nxt.error
import nxt.motor


@pytest.fixture
def sock():
    return Mock(spec_set=("send", "recv", "close"))


@pytest.fixture
def brick(sock):
    """A brick with a mock socket."""
    return nxt.brick.Brick(sock)


def sent(sent):
    return [call.send(sent)]


def sent_recved(sent):
    return [call.send(sent), call.recv()]


test_rxe_bin = b"test.rxe\0\0\0\0\0\0\0\0\0\0\0\0"
test_rso_bin = b"test.rso\0\0\0\0\0\0\0\0\0\0\0\0"
star_star_bin = b"*.*\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
loader_bin = b"Loader\0\0\0\0\0\0\0\0\0\0\0\0\0\0"


def test_close(sock, brick):
    brick.close()
    assert sock.mock_calls == [call.close()]


def test_with(sock, brick):
    with brick:
        pass
    assert sock.mock_calls == [call.close()]


def test_reply_error(sock, brick):
    sock.recv.return_value = bytes.fromhex("019700") + b"No\0"
    with pytest.raises(nxt.error.ProtocolError):
        brick.boot(sure=True)
    sock.recv.return_value = bytes.fromhex("029900") + b"No\0"
    with pytest.raises(nxt.error.ProtocolError):
        brick.boot(sure=True)
    sock.recv.return_value = bytes.fromhex("029799") + b"No\0"
    with pytest.raises(nxt.error.ProtocolError):
        brick.boot(sure=True)


class TestSystem:
    """Test system commands."""

    def test_file_open_read(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028000 42 01020304")
        handle, size = brick.file_open_read("test.rxe")
        assert sock.mock_calls == sent_recved(bytes.fromhex("0180") + test_rxe_bin)
        assert handle == 0x42
        assert size == 0x04030201

    def test_file_open_read_fail(self, sock, brick):
        """Test command failure reported in status code."""
        sock.recv.return_value = bytes.fromhex("028087 00 00000000")
        with pytest.raises(nxt.error.FileNotFoundError):
            brick.file_open_read("unknown.rxe")

    def test_file_open_write(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028100 42")
        handle = brick.file_open_write("test.rxe", 0x04030201)
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("0181") + test_rxe_bin + bytes.fromhex("01020304")
        )
        assert handle == 0x42

    def test_file_read(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028200 42 0700 21222324252627")
        handle, data = brick.file_read(0x42, 7)
        # TODO: The file_read command will return a EOF error if end of file is reached
        # before the read can be finished, and size will still contain the requested
        # data length instead of the actual read length. This is not the classic
        # semantic of a Unix read. Avoid reading past the end of file.
        assert sock.mock_calls == sent_recved(bytes.fromhex("0182 42 0700"))
        assert handle == 0x42
        assert data == bytes.fromhex("21222324252627")

    def test_file_write(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028300 42 0700")
        handle, size = brick.file_write(0x42, bytes.fromhex("21222324252627"))
        assert sock.mock_calls == sent_recved(bytes.fromhex("0183 42 21222324252627"))
        assert handle == 0x42
        assert size == 7

    def test_file_close(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028400 42")
        handle = brick.file_close(0x42)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0184 42"))
        assert handle == 0x42

    def test_file_delete(self, sock, brick):
        sock.recv.return_value = (
            bytes.fromhex("028500") + b"test.rxe\0\0\0\0\0\0\0\0\0\0\0\0"
        )
        name = brick.file_delete("test.rxe")
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("0185") + b"test.rxe\0\0\0\0\0\0\0\0\0\0\0\0"
        )
        assert name == "test.rxe"

    def test_file_find_first(self, sock, brick):
        sock.recv.return_value = (
            bytes.fromhex("028600 42") + test_rxe_bin + bytes.fromhex("01020304")
        )
        handle, name, size = brick.file_find_first("*.*")
        assert sock.mock_calls == sent_recved(bytes.fromhex("0186") + star_star_bin)
        assert handle == 0x42
        assert name == "test.rxe"
        assert size == 0x04030201

    def test_file_find_next(self, sock, brick):
        sock.recv.return_value = (
            bytes.fromhex("028700 42") + test_rxe_bin + bytes.fromhex("01020304")
        )
        handle, name, size = brick.file_find_next(0x42)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0187 42"))
        assert handle == 0x42
        assert name == "test.rxe"
        assert size == 0x04030201

    def test_get_firmware_version(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028800 0102 0304")
        prot_version, fw_version = brick.get_firmware_version()
        assert sock.mock_calls == sent_recved(bytes.fromhex("0188"))
        assert prot_version == (2, 1)
        assert fw_version == (4, 3)

    def test_file_open_write_linear(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028900 42")
        handle = brick.file_open_write_linear("test.rxe", 0x04030201)
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("0189") + test_rxe_bin + bytes.fromhex("01020304")
        )
        assert handle == 0x42

    def test_file_open_write_data(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028b00 42")
        handle = brick.file_open_write_data("test.rxe", 0x04030201)
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("018b") + test_rxe_bin + bytes.fromhex("01020304")
        )
        assert handle == 0x42

    def test_file_open_append_data(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("028c00 42 01020304")
        handle, available_size = brick.file_open_append_data("test.rxe")
        assert sock.mock_calls == sent_recved(bytes.fromhex("018c") + test_rxe_bin)
        assert handle == 0x42
        assert available_size == 0x04030201

    def test_module_find_first(self, sock, brick):
        sock.recv.return_value = (
            bytes.fromhex("029000 42")
            + loader_bin
            + bytes.fromhex("01000900 00000000 0800")
        )
        handle, name, mod_id, mod_size, mod_iomap_size = brick.module_find_first("*.*")
        assert sock.mock_calls == sent_recved(bytes.fromhex("0190") + star_star_bin)
        assert handle == 0x42
        assert name == "Loader"
        assert mod_id == 0x00090001
        assert mod_size == 0
        assert mod_iomap_size == 8

    def test_module_find_next(self, sock, brick):
        sock.recv.return_value = (
            bytes.fromhex("029100 42")
            + loader_bin
            + bytes.fromhex("01000900 00000000 0800")
        )
        handle, name, mod_id, mod_size, mod_iomap_size = brick.module_find_next(0x42)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0191 42"))
        assert handle == 0x42
        assert name == "Loader"
        assert mod_id == 0x00090001
        assert mod_size == 0
        assert mod_iomap_size == 8

    def test_module_close(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("029200 42")
        handle = brick.module_close(0x42)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0192 42"))
        assert handle == 0x42

    def test_read_io_map(self, sock, brick):
        sock.recv.return_value = bytes.fromhex(
            "029400 01020304 0700 212223242526270000"
        )
        mod_id, data = brick.read_io_map(0x04030201, 0x3231, 0x4241)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0194 01020304 3132 4142"))
        assert mod_id == 0x04030201
        assert data == bytes.fromhex("21222324252627")

    def test_write_io_map(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("029500 01020304 0700")
        mod_id, size = brick.write_io_map(
            0x04030201, 0x3231, bytes.fromhex("21222324252627")
        )
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("0195 01020304 3132 0700 21222324252627")
        )
        assert mod_id == 0x04030201
        assert size == 7

    def test_boot_not_sure(self, sock, brick):
        with pytest.raises(ValueError):
            brick.boot()

    def test_boot(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("029700") + b"Yes\0"
        resp = brick.boot(sure=True)
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("0197") + b"Let's dance: SAMBA\0"
        )
        assert resp == b"Yes\0"

    def test_set_brick_name(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("029800")
        brick.set_brick_name("NXT")
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("0198") + b"NXT\0\0\0\0\0\0\0\0\0\0\0\0"
        )

    def test_set_brick_name_too_long(self, sock, brick):
        with pytest.raises(ValueError):
            brick.set_brick_name("NXT456789012345")

    def test_get_device_info(self, sock, brick):
        sock.recv.return_value = (
            bytes.fromhex("029b00")
            + b"NXT\0\0\0\0\0\0\0\0\0\0\0\0"
            + bytes.fromhex("01020304050600" "11121314" "21222324")
        )
        name, address, signal_strengths, user_flash = brick.get_device_info()
        assert sock.mock_calls == sent_recved(bytes.fromhex("019b"))
        assert name == "NXT"
        assert address == "01:02:03:04:05:06"
        assert signal_strengths == (0x11, 0x12, 0x13, 0x14)
        assert user_flash == 0x24232221

    def test_delete_user_flash(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("02a000")
        brick.delete_user_flash()
        assert sock.mock_calls == sent_recved(bytes.fromhex("01a0"))

    def test_poll_command_length(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("02a100 01 07")
        buf_num, size = brick.poll_command_length(1)
        assert sock.mock_calls == sent_recved(bytes.fromhex("01a1 01"))
        assert buf_num == 1
        assert size == 7

    def test_poll_command(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("02a200 01 07 212223242526270000")
        buf_num, command = brick.poll_command(1, 9)
        assert sock.mock_calls == sent_recved(bytes.fromhex("01a2 01 09"))
        assert buf_num == 1
        assert command == bytes.fromhex("21222324252627")

    def test_bluetooth_factory_reset(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("02a400")
        brick.bluetooth_factory_reset()
        assert sock.mock_calls == sent_recved(bytes.fromhex("01a4"))


class TestDirect:
    """Test direct commands."""

    def test_start_program(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020000")
        brick.start_program("test.rxe")
        assert sock.mock_calls == sent_recved(bytes.fromhex("0000") + test_rxe_bin)

    def test_stop_program(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020100")
        brick.stop_program()
        assert sock.mock_calls == sent_recved(bytes.fromhex("0001"))

    def test_play_sound_file(self, sock, brick):
        brick.play_sound_file(True, "test.rso")
        assert sock.mock_calls == sent(bytes.fromhex("8002 01") + test_rso_bin)

    def test_play_tone(self, sock, brick):
        brick.play_tone(440, 1000)
        assert sock.mock_calls == sent(bytes.fromhex("8003 b801 e803"))

    def test_play_tone_and_wait(self, sock, brick):
        with patch("nxt.brick.time.sleep") as sleep:
            brick.play_tone_and_wait(440, 1000)
            assert sock.mock_calls == sent(bytes.fromhex("8003 b801 e803"))
            assert sleep.mock_calls == [call(1)]

    def test_set_output_state(self, sock, brick):
        brick.set_output_state(
            nxt.motor.Port.B,
            -100,
            nxt.motor.Mode.ON,
            nxt.motor.RegulationMode.IDLE,
            -5,
            nxt.motor.RunState.RUNNING,
            0x04030201,
        )
        assert sock.mock_calls == sent(bytes.fromhex("8004 01 9c 01 00 fb 20 01020304"))

    def test_set_input_mode(self, sock, brick):
        brick.set_input_mode(
            nxt.sensor.Port.S3, nxt.sensor.Type.SWITCH, nxt.sensor.Mode.BOOL
        )
        assert sock.mock_calls == sent(bytes.fromhex("8005 02 01 20"))

    def test_get_output_state(self, sock, brick):
        sock.recv.return_value = bytes.fromhex(
            "020600 01 9c 01 00 fb 20 01020304 11121314 21222324 31323334"
        )
        (
            port,
            power,
            mode,
            regulation_mode,
            turn_ratio,
            run_state,
            tacho_limit,
            tacho_count,
            block_tacho_count,
            rotation_count,
        ) = brick.get_output_state(nxt.motor.Port.B)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0006 01"))
        assert port == nxt.motor.Port.B
        assert power == -100
        assert mode == nxt.motor.Mode.ON
        assert regulation_mode == nxt.motor.RegulationMode.IDLE
        assert turn_ratio == -5
        assert run_state == nxt.motor.RunState.RUNNING
        assert tacho_limit == 0x04030201
        assert tacho_count == 0x14131211
        assert block_tacho_count == 0x24232221
        assert rotation_count == 0x34333231

    def test_get_input_values(self, sock, brick):
        sock.recv.return_value = bytes.fromhex(
            "020700 02 01 00 01 20 0102 1112 2122 3132"
        )
        (
            port,
            valid,
            calibrated,
            sensor_type,
            sensor_mode,
            raw_value,
            normalized_value,
            scaled_value,
            calibrated_value,
        ) = brick.get_input_values(nxt.sensor.Port.S3)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0007 02"))
        assert port == nxt.sensor.Port.S3
        assert valid is True
        assert calibrated is False
        assert sensor_type == nxt.sensor.Type.SWITCH
        assert sensor_mode == nxt.sensor.Mode.BOOL
        assert raw_value == 0x0201
        assert normalized_value == 0x1211
        assert scaled_value == 0x2221
        assert calibrated_value == 0x3231

    def test_reset_input_scaled_value(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020800")
        brick.reset_input_scaled_value(nxt.sensor.Port.S3)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0008 02"))

    def test_message_write(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020900")
        brick.message_write(3, bytes.fromhex("21222324252627"))
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("0009 03 08 21222324252627 00")
        )

    def test_message_write_too_large(self, sock, brick):
        with pytest.raises(ValueError):
            brick.message_write(3, bytes.fromhex("21") * 59)

    def test_reset_motor_position(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020a00")
        brick.reset_motor_position(nxt.motor.Port.B, True)
        assert sock.mock_calls == sent_recved(bytes.fromhex("000a 01 01"))

    def test_get_battery_level(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020b00 2823")
        millivolts = brick.get_battery_level()
        assert sock.mock_calls == sent_recved(bytes.fromhex("000b"))
        assert millivolts == 9000

    def test_stop_sound_playback(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020c00")
        brick.stop_sound_playback()
        assert sock.mock_calls == sent_recved(bytes.fromhex("000c"))

    def test_keep_alive(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020d00 01020304")
        sleep_timeout = brick.keep_alive()
        assert sock.mock_calls == sent_recved(bytes.fromhex("000d"))
        assert sleep_timeout == 0x04030201

    def test_ls_get_status(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020e00 07")
        size = brick.ls_get_status(nxt.sensor.Port.S3)
        assert sock.mock_calls == sent_recved(bytes.fromhex("000e 02"))
        assert size == 7

    def test_ls_write(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("020f00")
        brick.ls_write(nxt.sensor.Port.S3, bytes.fromhex("21222324252627"), 9)
        assert sock.mock_calls == sent_recved(
            bytes.fromhex("000f 02 07 09 21222324252627")
        )

    def test_ls_read(self, sock, brick):
        sock.recv.return_value = bytes.fromhex(
            "021000 07 21222324252627 000000000000000000"
        )
        rx_data = brick.ls_read(nxt.sensor.Port.S3)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0010 02"))
        assert rx_data == bytes.fromhex("21222324252627")

    def test_get_current_program_name(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("021100") + test_rxe_bin
        name = brick.get_current_program_name()
        assert sock.mock_calls == sent_recved(bytes.fromhex("0011"))
        assert name == "test.rxe"

    def test_message_read(self, sock, brick):
        sock.recv.return_value = bytes.fromhex("021300 00 07 21222324252627") + bytes(
            52
        )
        local_inbox, message = brick.message_read(10, 0, True)
        assert sock.mock_calls == sent_recved(bytes.fromhex("0013 0a 00 01"))
        assert local_inbox == 0
        assert message == bytes.fromhex("21222324252627")


class TestFilesModules:
    """Test nxt.brick files & modules access."""

    def test_find_files_none(self, mbrick):
        mbrick.file_find_first.side_effect = [nxt.error.FileNotFoundError()]
        results = list(mbrick.find_files("*.*"))
        assert results == []
        assert mbrick.mock_calls == [
            call.file_find_first("*.*"),
        ]

    def test_find_files_one(self, mbrick):
        mbrick.file_find_first.return_value = (0x42, "test.rxe", 0x04030201)
        mbrick.file_find_next.side_effect = [nxt.error.FileNotFoundError()]
        results = list(mbrick.find_files("*.*"))
        assert results == [("test.rxe", 0x04030201)]
        assert mbrick.mock_calls == [
            call.file_find_first("*.*"),
            call.file_find_next(0x42),
            call.file_close(0x42),
        ]

    def test_find_files_two(self, mbrick):
        mbrick.file_find_first.return_value = (0x42, "test.rxe", 0x04030201)
        mbrick.file_find_next.side_effect = [
            (0x42, "test.rso", 7),
            nxt.error.FileNotFoundError(),
        ]
        results = list(mbrick.find_files("*.*"))
        assert results == [("test.rxe", 0x04030201), ("test.rso", 7)]
        assert mbrick.mock_calls == [
            call.file_find_first("*.*"),
            call.file_find_next(0x42),
            call.file_find_next(0x42),
            call.file_close(0x42),
        ]

    def test_find_files_interrupted(self, mbrick):
        mbrick.file_find_first.return_value = (0x42, "test.rxe", 0x04030201)
        mbrick.file_find_next.side_effect = [nxt.error.FileNotFoundError()]
        g = mbrick.find_files("*.*")
        assert next(g) == ("test.rxe", 0x04030201)
        g.close()
        assert mbrick.mock_calls == [
            call.file_find_first("*.*"),
            call.file_close(0x42),
        ]

    def test_find_modules_none(self, mbrick):
        mbrick.module_find_first.side_effect = [nxt.error.ModuleNotFoundError()]
        results = list(mbrick.find_modules("*.*"))
        assert results == []
        assert mbrick.mock_calls == [
            call.module_find_first("*.*"),
        ]

    def test_find_modules_one(self, mbrick):
        mbrick.module_find_first.return_value = (0x42, "Loader", 0x00090001, 0, 8)
        mbrick.module_find_next.side_effect = [nxt.error.ModuleNotFoundError()]
        results = list(mbrick.find_modules("*.*"))
        assert mbrick.mock_calls == [
            call.module_find_first("*.*"),
            call.module_find_next(0x42),
            call.module_close(0x42),
        ]
        assert results == [("Loader", 0x00090001, 0, 8)]

    def test_find_modules_two(self, mbrick):
        mbrick.module_find_first.return_value = (0x42, "Loader", 0x00090001, 0, 8)
        mbrick.module_find_next.side_effect = [
            (0x42, "Dummy", 0x01020304, 0, 12),
            nxt.error.ModuleNotFoundError(),
        ]
        results = list(mbrick.find_modules("*.*"))
        assert mbrick.mock_calls == [
            call.module_find_first("*.*"),
            call.module_find_next(0x42),
            call.module_find_next(0x42),
            call.module_close(0x42),
        ]
        assert results == [
            ("Loader", 0x00090001, 0, 8),
            ("Dummy", 0x01020304, 0, 12),
        ]

    def test_file_read_text(self, mbrick):
        mbrick.file_open_read.return_value = (0x42, 12)
        mbrick.file_read.return_value = (0x42, b"hello\nworld\n")
        with mbrick.open_file("test.txt") as f:
            results = list(f)
        assert results == ["hello\n", "world\n"]
        assert mbrick.mock_calls == [
            call.file_open_read("test.txt"),
            call.file_read(0x42, 12),
            call.file_close(0x42),
        ]

    def test_file_read_bin(self, mbrick):
        mbrick.file_open_read.return_value = (0x42, 7)
        mbrick.file_read.return_value = (0x42, bytes.fromhex("21222324252627"))
        with mbrick.open_file("test.bin", "rb") as f:
            assert f.read() == bytes.fromhex("21222324252627")
        assert mbrick.mock_calls == [
            call.file_open_read("test.bin"),
            call.file_read(0x42, 7),
            call.file_close(0x42),
        ]

    def test_file_read_raw(self, mbrick):
        mbrick.file_open_read.return_value = (0x42, 7)
        mbrick.file_read.return_value = (0x42, bytes.fromhex("21222324252627"))
        with mbrick.open_file("test.bin", "rb", buffering=0) as f:
            assert f.read() == bytes.fromhex("21222324252627")
        assert mbrick.mock_calls == [
            call.file_open_read("test.bin"),
            call.file_read(0x42, 7),
            call.file_close(0x42),
        ]

    def test_file_write_text(self, mbrick):
        mbrick.file_open_write.return_value = 0x42
        mbrick.file_write.return_value = (0x42, 12)
        f = mbrick.open_file("test.txt", "wt", 12)
        f.write("hello\n")
        f.write("world\n")
        f.close()
        assert mbrick.mock_calls == [
            call.file_open_write("test.txt", 12),
            call.file_write(0x42, b"hello\nworld\n"),
            call.file_close(0x42),
        ]

    def test_file_write_text_encoding(self, mbrick):
        mbrick.file_open_write.return_value = 0x42
        mbrick.file_write.return_value = (0x42, 16)
        f = mbrick.open_file("test.txt", "wt", 16, errors="replace")
        f.write("un petit café ?\n")
        f.close()
        assert mbrick.mock_calls == [
            call.file_open_write("test.txt", 16),
            call.file_write(0x42, b"un petit caf? ?\n"),
            call.file_close(0x42),
        ]

    def test_file_write_bin(self, mbrick):
        mbrick.file_open_write.return_value = 0x42
        mbrick.file_write.return_value = (0x42, 7)
        f = mbrick.open_file("test.bin", "wb", 7)
        f.write(bytes.fromhex("21222324252627"))
        f.close()
        assert mbrick.mock_calls == [
            call.file_open_write("test.bin", 7),
            call.file_write(0x42, bytes.fromhex("21222324252627")),
            call.file_close(0x42),
        ]

    def test_file_write_raw(self, mbrick):
        mbrick.file_open_write.return_value = 0x42
        mbrick.file_write.return_value = (0x42, 7)
        f = mbrick.open_file("test.bin", "wb", 7, buffering=0)
        f.write(bytes.fromhex("21222324252627"))
        f.close()
        assert mbrick.mock_calls == [
            call.file_open_write("test.bin", 7),
            call.file_write(0x42, bytes.fromhex("21222324252627")),
            call.file_close(0x42),
        ]

    def test_file_write_too_much(self, mbrick):
        mbrick.file_open_write.return_value = 0x42
        mbrick.file_write.return_value = (0x42, 7)
        f = mbrick.open_file("test.bin", "wb", 7)
        f.write(bytes.fromhex("2122232425262728"))
        with pytest.raises(ValueError):
            # Flush done on close.
            f.close()
        assert f.closed
        assert mbrick.mock_calls == [
            call.file_open_write("test.bin", 7),
            call.file_write(0x42, bytes.fromhex("21222324252627")),
            call.file_close(0x42),
        ]

    def test_file_write_closed(self, mbrick):
        mbrick.file_open_write.return_value = 0x42
        mbrick.file_write.return_value = (0x42, 7)
        # Write directly to raw file to bypass error check in buffered file.
        f = mbrick.open_file("test.bin", "wb", 7, buffering=0)
        f.close()
        assert f.closed
        with pytest.raises(ValueError):
            f.write(bytes.fromhex("28"))
        assert mbrick.mock_calls == [
            call.file_open_write("test.bin", 7),
            call.file_close(0x42),
        ]

    def test_file_invalid_params(self, mbrick):
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "ww", 7)
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "wr", 7)
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "tb")
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "t")
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "x")
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "rb", encoding="ascii")
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "rb", errors="ignore")
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "rb", newline="\n")
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "r", buffering=0)
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "r", 7)
        with pytest.raises(ValueError):
            mbrick.open_file("test.bin", "w")



#!/usr/bin/python3
"""NXT-Python tutorial: use touch sensor."""
import time

import nxt.locator
import nxt.sensor
import nxt.sensor.generic

with nxt.locator.find() as b:
    # Get the sensor connected to port 1, not a digital sensor, must give the sensor
    # class.
    mysensor = b.get_sensor(nxt.sensor.Port.S1, nxt.sensor.generic.Touch)
    # Read the sensor in a loop (until interrupted).
    print("Use Ctrl-C to interrupt")
    while True:
        value = mysensor.get_sample()
        print(value)
        time.sleep(0.5)



# nxt.sensor.mindsensors module -- Classes implementing Mindsensors sensors
# Copyright (C) 2006,2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner, Paulo Vieira, rhn
# Copyright (C) 2010  Marcus Wanner, MindSensors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import logging

from nxt.sensor import Type, Mode
from .digital import BaseDigitalSensor, SensorInfo
from .analog import BaseAnalogSensor

logger = logging.getLogger(__name__)

class SumoEyes(BaseAnalogSensor):   
    """The class to control Mindsensors Sumo sensor. Warning: long range not
    working for my sensor.
    """
    #range: 5-10cm
    class Reading:
        """Contains the reading of SumoEyes sensor. left and right can be True or
        False. If True, then there is something there, if False, then it's empty
        there.
        """
        def __init__(self, raw_reading):
            self.raw = raw_reading
            val = raw_reading.normalized_value # FIXME: make it rely on raw_value
            right = 600 < val < 700
            both = 700 <= val < 900
            left = 300 < val < 400
            self.left = left or both
            self.right = right or both
        
        def __str__(self):
            return '(left: ' + str(self.left) + ', right: ' + str(self.right) + ')'

    def __init__(self, brick, port, long_range=False):
        super().__init__(brick, port)
        self.set_long_range(long_range)
        
    def set_long_range(self, val):
        """Sets if the sensor should operate in long range mode (12 inches) or
        the short range mode (6 in). val should be True or False.
        """
        if val:
            type_ = Type.LIGHT_INACTIVE
        else:
            type_ = Type.LIGHT_ACTIVE
        self.set_input_mode(type_, Mode.RAW)
    
    def get_sample(self):
        """Returns the processed meaningful values of the sensor"""
        return self.Reading(self.get_input_values())


class Compassv2(BaseDigitalSensor):
    """Class for the now-discontinued CMPS-Nx sensor. Also works with v1.1 sensors.
Note that when using a v1.x sensor, some of the commands are not supported!
To determine your sensor's version, use get_sensor_info().version"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'command': (0x41, '<B'),
                        'heading': (0x42, '<H'),
                        'x_offset': (0x44, '<h'), #unsure about signedness for this one
                        'y_offset': (0x46, '<h'), #and this one
                        'x_range': (0x48, '<H'),
                        'y_range': (0x4A, '<H'),
                        'x_raw': (0x4C, '<H'), #and this one
                        'y_raw': (0x4E, '<H'), #and this one
    })
    
    class Commands:
        AUTO_TRIG_ON = 'A'
        AUTO_TRIG_OFF = 'S'
        MAP_HEADING_BYTE = 'B'      # map heading to 0-255 range
        MAP_HEADING_INTEGER = 'I'   # map heading to 0-36000 (or 3600) range
        SAMPLING_50_HZ = 'E'        # set sampling frequency to 50 Hz
        SAMPLING_60_HZ = 'U'        # set sampling frequency to 60 Hz
        SET_ADPA_MODE_ON = 'N'      # set ADPA mode on
        SET_ADPA_MODE_OFF = 'O'     # set ADPA mode off
        BEGIN_CALIBRATION = 'C'     # begin calibration
        DONE_CALIBRATION = 'D'      # done with calibration
        LOAD_USER_CALIBRATION = 'L' # load user calibration value
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)
        self.command(self.Commands.MAP_HEADING_INTEGER)

    def command(self, command):
        value = ord(command)
        self.write_value('command', (value, ))
    
    def get_heading(self):
        return self.read_value('heading')[0]
        
    get_sample = get_heading

Compassv2.add_compatible_sensor(None, 'mndsnsrs', 'CMPS')


class DIST(BaseDigitalSensor):  
    """Class for the Distance Infrared Sensor"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'command': (0x41, '<B'),
                        'distance': (0x42, '<H'),
                        'voltage': (0x44, '<H'),
                        'type': (0x50, '<B'),
                        'no_of_data_points': (0x51, '<B'),
                        'min_distance': (0x52, '<H'),
                        'max_distance': (0x54, '<H'),
    })

    class Commands:
        TYPE_GP2D12 = '1' #GP2D12 sensor Module
        TYPE_GP2D120 = '2' #Short range sensor Module
        TYPE_GP2Y0A21YK = '3' #Medium range sensor Module
        TYPE_GP2Y0A02YK = '4' #Long range sensor Module
        TYPE_CUSTOM = '5' #Custom sensor Module
        POWER_ON = 'E' #Sensor module power on
        POWER_OFF = 'D' #Sensor module power offset
        ADPA_ON = 'N' #ADPA mode on
        ADPA_OFF = 'O' #ADPA mode off (default)
    
    def command(self, command):
        value = ord(command)
        self.write_value('command', (value, ))
    
    def get_distance(self):
        return self.read_value('distance')[0]

    get_sample = get_distance

    def get_type(self):
        return self.read_value('type')[0]
        
    def get_voltage(self):
        return self.read_value('voltage')[0]
    
    def get_min_distance(self):
        return self.read_value('min_distance')[0]
        
    def get_max_distance(self):
        return self.read_value('max_distance')[0]

DIST.add_compatible_sensor(None, 'mndsnsrs', 'DIST')


class RTC(BaseDigitalSensor):
    """Class for the RealTime Clock sensor (DS1307)"""
    #TODO: Create a function to set the clock
    #Has no identification
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'seconds': (0x00, '<B'),
                        'minutes': (0x01, '<B'),
                        'hours':   (0x02, '<B'),
                        'day':     (0x03, '<B'),
                        'date':    (0x04, '<B'),
                        'month':   (0x05, '<B'),
                        'year':    (0x06, '<B'),
                      })
    I2C_DEV = 0xD0

    def __init__(self, brick, port):
        super().__init__(brick, port, check_compatible=False)

    def get_seconds(self):  
        gs = self.read_value('seconds')[0]
        gs2 = gs & 0xf  # bitmasks
        gs3 = gs & 0x70
        gs3 = gs3 >> 4
        return str(gs3) + str(gs2)
    
    def get_minutes(self):
        gm = self.read_value('minutes')[0]
        gm2 = gm & 0xf
        gm3 = gm & 0x70
        gm3 = gm3 >> 4
        return str(gm3) + str(gm2)
        
    def get_hours(self):    
        gh = self.read_value('hours')[0]
        gh2 = gh & 0xf
        gh3 = gh & 0x30
        gh3 = gh3 >> 4
        return str(gh3) + str(gh2)

    def get_day(self):      
        gwd = self.read_value('day')[0]
        gwd = gwd & 0x07
        return gwd

    def get_month(self):    
        gmo = self.read_value('month')[0]
        gmo2 = gmo & 0xf
        gmo3 = gmo & 0x10
        gmo3 = gmo3 >> 4
        return str(gmo3) + str(gmo2)

    def get_year(self):
        """Last two digits (10 for 2010)"""
        gy = self.read_value('year')[0]
        gy2 = gy & 0xf
        gy3 = gy & 0xF0
        gy3 =  gy3 >> 4
        return str(gy3) + str(gy2)

    def get_date(self):
        gd = self.read_value('date')[0]
        gd2 = gd & 0xf
        gd3 = gd & 0x60
        gd3 = gd3 >> 4
        return str(gd3) + str(gd2)

    def hour_mode(self, mode):
        """Writes mode bit and re-enters hours, which is required"""
        if mode == 12 or 24:
            hm = self.read_value('hours')[0]
            hm2 = hm & 0x40
            hm2 = hm2 >> 6
            if mode == 12 and hm2 == 0:  #12_HOUR = 1
                hm3 = hm + 64
                self.write_value('hours', (hm3, ))
            elif mode == 24 and hm2 == 1:  #24_HOUR = 0
                hm3 = hm - 64
                self.write_value('hours', (hm3, ))
        else:
            raise ValueError('Must be 12 or 24!')

    def get_mer(self):
        mer = self.read_value('hours')[0]
        mer2 = mer & 0x40
        mer2 = mer2 >> 6
        if mer2 == 1:
            mer3 = mer & 0x20
            mer3 = mer3 >> 0x10
            return mer3
        else:
            logger.error('cannot get mer in 24-hour mode')
    
    def get_sample(self):
        """Returns a struct_time() tuple which can be processed by the time module."""
        import time
        return time.struct_time((
            int(self.get_year())+2000,
            int(self.get_month()),
            int(self.get_date()),
            int(self.get_hours()),
            int(self.get_minutes()),
            int(self.get_seconds()),
            int(self.get_day()),
            0, #Should be the Julian Day, but computing that is hard.
            0 #No daylight savings time to worry about here.
            ))


class ACCL(BaseDigitalSensor):
    """Class for Accelerometer sensor"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'sensitivity':  (0x19, 'B'),
                        'command':      (0x41, 'B'),
                        'x_tilt':       (0x42, 'b'),
                        'y_tilt':       (0x43, 'b'),
                        'z_tilt':       (0x44, 'b'),
                        'all_tilt':     (0x42, '3b'),

                        'x_accel':      (0x45, '<h'),
                        'y_accel':      (0x47, '<h'),
                        'z_accel':      (0x49, '<h'),
                        'all_accel':    (0x45, '<3h'),

                        'x_offset': (0x4B, '<h'),
                        'x_range':  (0x4D, '<h'),

                        'y_offset': (0x4F, '<h'),
                        'y_range':  (0x51, '<h'),

                        'z_offset': (0x53, '<h'),
                        'z_range':  (0x55, '<h'),
                      })
    
    class Commands:
        SENS_15G = '1' #that's 1.5...Alt. 2.5G (sensors older than V3.20)
        SENS_2G = '2' #Alt .3.3G
        SENS_4G = '3' #Alt. 6.7G
        SENS_6G = '4' #Alt. 10G
        X_CALIBRATION = 'X' #Acquire X point calibration
        X_CAL_AND_END = 'x' #X point calibration and end calibration
        Y_CALIBRATION = 'Y' #Acquire Y point calibration
        Y_CAL_AND_END = 'y' #Y point calibration and end calibration
        Z_CALIBRATION = 'Z' #Acquire Z point calibration
        Z_CAL_AND_END = 'z' #Z point calibration and end calibration
        CAL_RESET = 'R' #Reset to factory set calibration
        ADPA_ON = 'N' #Set ADPA mode On
        ADPA_OFF = 'O' #Set ADPA mode Off (default)
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def command(self, command):
        value = ord(command)
        self.write_value('command', (value, ))
    
    def get_sensitivity(self):
        return chr(self.read_value('sensitivity')[0])

    def get_tilt(self, axis):
        xyz = str(axis) + '_tilt'
        return self.read_value(xyz)[0]
    
    def get_all_tilt(self):
        return self.read_value('all_tilt')
    
    def get_accel(self, axis):
        xyz = str(axis) + '_accel'
        return self.read_value(xyz)[0]
    
    def get_all_accel(self):
        return self.read_value('all_accel')
    
    get_sample = get_all_accel

    def get_offset(self, axis):
        xyz = str(axis) + '_offset'
        return self.read_value(xyz)[0]

    def get_range(self, axis):
        xyz = str(axis) + '_range'
        return self.read_value(xyz)[0]

    def set_offset(self, axis, value):
        xyz = str(axis) + '_offset'
        self.write_value(xyz, (value, ))

    def set_range(self, axis, value):
        xyz = str(axis) + '_range'
        self.write_value(xyz, (value, ))

ACCL.add_compatible_sensor(None, 'mndsnsrs', 'ACCL-NX') #Tested with version 'V3.20'


class MTRMUX(BaseDigitalSensor):
    """Class for RCX Motor Multiplexer sensor"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'command'     :  (0x41, '<B'),
                        'direction_m1':  (0x42, '<B'),
                        'speed_m1':      (0x43, '<B'),
                        'direction_m2':  (0x44, '<B'),
                        'speed_m2':      (0x45, '<B'),
                        'direction_m3':  (0x46, '<B'),
                        'speed_m3':      (0x47, '<B'),
                        'direction_m4':  (0x48, '<B'),
                        'speed_m4':      (0x49, '<B'),
                      })
    I2C_DEV = 0xB4
    
    class Commands:
        FLOAT = 0x00
        FORWARD = 0x01
        REVERSE = 0x02
        BRAKE = 0x03    
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def command(self, command):
        self.write_value('command', (command, )) 

    def set_direction(self, number, value):
        addressname = 'direction_m' + str(number)
        self.write_value(addressname, (value, ))

    def set_speed(self, number, value):
        addressname = 'speed_m' + str(number)
        self.write_value(addressname, (value, ))

    def get_direction(self, number):
        addressname = 'direction_m' + str(number)
        return self.read_value(addressname)[0]

    def get_speed(self, number):
        addressname = 'speed_m' + str(number)
        return self.read_value(addressname)[0]

MTRMUX.add_compatible_sensor(None, 'mndsnsrs', 'MTRMUX') #Tested with version 'V2.11'


class LineLeader(BaseDigitalSensor):
    """Class for Line Sensor Array"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'command':   (0x41, '<B'),
                        'steering':  (0x42, '<b'),
                        'average':   (0x43, '<B'),
                        'result':    (0x44, '<B'),
                        'set_point': (0x45, '<B'),
                
                        'kp':        (0x46, '<B'),
                        'ki':        (0x47, '<B'),
                        'kd':        (0x48, '<B'),
                        'kp_divisor':(0x61, '<B'),
                        'ki_divisor':(0x62, '<B'),
                        'kd_divisor':(0x63, '<B'),
                 #One byte for each sensor, so byte# = sensor#
                        'calibrated_reading_byte1':  (0x49, '<B'),
                        'calibrated_reading_byte2':  (0x4A, '<B'),
                        'calibrated_reading_byte3':  (0x4B, '<B'),
                        'calibrated_reading_byte4':  (0x4C, '<B'),
                        'calibrated_reading_byte5':  (0x4D, '<B'),
                        'calibrated_reading_byte6':  (0x4E, '<B'),
                        'calibrated_reading_byte7':  (0x4F, '<B'),
                        'calibrated_reading_byte8':  (0x50, '<B'),
                        'all_calibrated_readings':   (0x49, '<8B'),

                        'w_read_limit':(0x51, '<H'),
                        'b_read_limit':(0x59, '<B'),
                        'w_cal_data1':(0x64, '<B'),
                        'b_cal_data':(0x6C, '<B'),

                        'uncal_sensor1_voltage_byte1':(0x74, '<B'),
                        'uncal_sensor2_voltage_byte1':(0x76, '<B'),
                        'uncal_sensor3_voltage_byte1':(0x78, '<B'),
                        'uncal_sensor4_voltage_byte1':(0x7A, '<B'),
                        'uncal_sensor5_voltage_byte1':(0x7C, '<B'),
                        'uncal_sensor6_voltage_byte1':(0x7E, '<B'),
                        'uncal_sensor7_voltage_byte1':(0x80, '<B'),
                        'uncal_sensor8_voltage_byte1':(0x82, '<B'),
                        'all_uncal_readings':         (0x74, '<8B'),
                      })
    
    class Commands:
        CALIBRATE_WHITE = 'W'
        CALIBRATE_BLACK = 'B'
        SENSOR_SLEEP = 'D'
        US_CONFIG = 'A'
        EU_CONFIG = 'E'
        UNI_CONFIG = 'U'
        SENSOR_WAKE = 'P'
        COLOR_INVERT = 'I'
        COLOR_INVERT_REVERSE = 'R'
        SNAPSHOT = 'S'
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def command(self, command):
        value = ord(command)
        self.write_value('command', (value, ))

    def get_steering(self):
        'Value to add to the left and subtract from the right motor\'s power.'
        return self.read_value('steering')[0]

    def get_average(self):
        'Weighted average; greater as line is closer to right edge. 0 for no line.'
        return self.read_value('average')[0]

    def get_result(self):
        'Bitmap, one bit for each sensor'
        return self.read_value('result')[0]

    def set_set_point(self, value):
        'Average value for steering to gravitate to. 10 (left) to 80 (right).'
        self.write_value('set_point', (value, ))

    def set_pid(self, pid, value):
        addressname = 'k' + str(pid)
        self.write_value(addressname, (value, ))

    def set_pid_divisor(self, pid, value):
        addressname = 'k' + str(pid) +  '_divisor'
        self.write_value(addressname, (value, ))

    def get_reading(self, number):
        addressname = 'calibrated_reading_byte' + str(number)
        return self.read_value(addressname)[0]
    
    def get_reading_all(self):
        return self.read_value('all_calibrated_readings')
    
    get_sample = get_reading_all
    
    def get_uncal_reading(self, number):
        addressname = 'uncal_sensor' + str(number) + '_voltage_byte1'
        return self.read_value(addressname)[0]
    
    def get_uncal_all(self):
        return self.read_value('all_uncal_readings')

LineLeader.add_compatible_sensor(None, 'mndsnsrs', 'LineLdr') #Tested with version 'V1.16'


class Servo(BaseDigitalSensor):
    """Class for Servo sensors"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'command'  : (0x41, '<B'),

                        'servo_1_pos':  (0x42, '<H'),
                        'servo_2_pos':  (0x44, '<H'),
                        'servo_3_pos':  (0x46, '<H'),
                        'servo_4_pos':  (0x48, '<H'),
                        'servo_5_pos':  (0x4A, '<H'),
                        'servo_6_pos':  (0x4C, '<H'),
                        'servo_7_pos':  (0x4E, '<H'),
                        'servo_8_pos':  (0x50, '<H'),

                        'servo_1_speed':    (0x52, '<B'),
                        'servo_2_speed':    (0x53, '<B'),
                        'servo_3_speed':    (0x54, '<B'),
                        'servo_4_speed':    (0x55, '<B'),
                        'servo_5_speed':    (0x56, '<B'),
                        'servo_6_speed':    (0x57, '<B'),
                        'servo_7_speed':    (0x58, '<B'),
                        'servo_8_speed':    (0x59, '<B'),
                  
                        'servo_1_quick':      (0x5A, '<B'),
                        'servo_2_quick':      (0x5B, '<B'),
                        'servo_3_quick':      (0x5C, '<B'),
                        'servo_4_quick':      (0x5D, '<B'),
                        'servo_5_quick':      (0x5E, '<B'),
                        'servo_6_quick':      (0x5F, '<B'),
                        'servo_7_quick':      (0x60, '<B'),
                        'servo_8_quick':      (0x61, '<B'),
                     })
    I2C_DEV = 0xB0
    
    COMMANDVALUES = {'R':  (0x52),   #Resume macro execution
                'S':  (0x53),   #reset initial position and speed
                'I1': (0x4931), #store initial position motor 1
                'I2': (0x4932), #store initial position motor 2
                'I3': (0x4933), #etc...
                'I4': (0x4934),
                'I5': (0x4935),
                'I6': (0x4936),
                'I7': (0x4937),
                'I8': (0x4938),
                'H':  (0x48),   #Halt macro
                'Gx': (0x4778), #not going to work yet x = variable
                'EM': (0x454d), #Edit Macro
                'P':  (0x50),   #Pause Macro
                }
    
    class Commands:
        RESUME_MACRO = 'R'
        RESET_POS_SPEED = 'S'
        STORE_MOTOR_POS_1 = 'I1'
        STORE_MOTOR_POS_2 = 'I2'
        STORE_MOTOR_POS_3 = 'I3'
        STORE_MOTOR_POS_4 = 'I4'
        STORE_MOTOR_POS_5 = 'I5'
        STORE_MOTOR_POS_6 = 'I6'
        STORE_MOTOR_POS_7 = 'I7'
        STORE_MOTOR_POS_8 = 'I8'
        HALT_MACRO = 'H'
        X_TO_VAR = 'Gx' #not going to work yet
        EDIT_MACRO = 'EM'
        PAUSE_MACRO = 'P'
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)
        
    def command(self, command):
        value = self.COMMANDVALUES[command]
        self.write_value('command', (value, ))
    
    def get_bat_level(self):
        return self.read_value('command')[0]

    def set_position(self, number, value):
       addressname = 'servo_' + str(number) + '_pos'
       self.write_value(addressname, (value, ))
    
    def get_position(self, number):
        return self.read_value('servo_' + str(number) + '_pos')[0]
    
    def set_speed(self, number, value):
       addressname = 'servo_' + str(number) + '_speed'
       self.write_value(addressname, (value, ))
    
    def get_speed(self, number):
        return self.read_value('servo_' + str(number) + '_speed')[0]
    
    def set_quick(self, number, value):
           addressname = 'servo_' + str(number) + '_quick'
           self.write_value(addressname, (value, ))
            
Servo.add_compatible_sensor(None, 'mndsnsrs', 'NXTServo') #Tested with version 'V1.20'


class MMX(BaseDigitalSensor):
    """Class for MMX sensors"""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'command'  :        (0x41, '<B'),
                        #Motor Writes
                        'encoder_1_target': (0x42, '<l'),
                        'speed_1':          (0x46, '<B'),
                        'seconds_to_run_1': (0x47, '<B'),
                        'command_b_1':      (0x48, '<B'),
                        'command_a_1':      (0x49, '<B'),

                        'encoder_2_target': (0x4A, '<l'),
                        'speed_2':          (0x4E, '<B'),
                        'seconds_to_run_2': (0x4F, '<B'),
                        'command_b_2':      (0x50, '<B'),
                        'command_a_2':      (0x51, '<B'),
                        #Motor reads
                        'encoder_1_pos':    (0x62, '<H'),
                        'encoder_2_pos':    (0x66, '<H'),
                        'status_m1':        (0x72, '<B'),
                        'status_m2':        (0x73, '<B'),
                        'tasks_running_m1': (0x76, '<H'),
                        'tasks_running_m2': (0x77, '<H'),
                        #PID Control
                        'p_encoder':       (0x7A, '<H'),
                        'i_encoder':       (0x7C, '<H'),
                        'd_encoder':       (0x7E, '<H'),
                        'p_speed':         (0x80, '<H'),
                        'i_speed':         (0x82, '<H'),
                        'd_speed':         (0x84, '<H'),
                        'pass_count':       (0x86, '<B'),
                        'tolerance':        (0x87, '<B'),
                  })      
    I2C_DEV = 0x06
    
    class Commands:
        RESET_PARAMS_ENCODERS = 'R'
        ISSUE_SYNCED_COMMANDS = 'S'
        MOTOR_1_FLOAT_STOP = 'a'
        MOTOR_2_FLOAT_STOP = 'b'
        BOTH_FLOAT_STOP = 'c'
        MOTOR_1_BRAKE_STOP = 'A'
        MOTOR_2_BRAKE_STOP = 'B'
        BOTH_BRAKE_STOP = 'C'
        MOTOR_1_ENC_RESET = 'r'
        MOTOR_2_ENC_RESET = 's'
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)
    
    def command(self, command):
        value = ord(command)
        self.write_value('command', (value, ))
    
    def get_bat_level(self):
        return self.read_value('command')[0]

    def set_encoder_target(self, motor_number, value):
        addressname = 'encoder_' + str(motor_number) + '_target'
        self.write_value(addressname, (value, ))

    def set_speed(self, motor_number, value):
        addressname = 'speed_' + str(motor_number)
        self.write_value(addressname, (value, ))

    def set_time_run(self, motor_number, seconds):
        addressname = 'seconds_to_run_' + str(motor_number)
        self.write_value(addressname, (seconds, ))

    def command_b(self, motor_number, value):
        addressname = 'command_b_' + str(motor_number)
        self.write_value(addressname, (value, ))

    def command_a(self, motor_number, bit_num, bit_val):
        addressname = 'command_a_' + str(motor_number)
        s = self.read_value(addressname)[0]
        #I feel like there must be an easier way to write one bit...
        val = bit_val << bit_num
        if bit_val == 1:    
            value = val | s
            self.write_value(addressname, (value, ))
            return value #testing purposes
        elif bit_val == 0:
            val = 1
            val = val << bit_num
            val = val ^ 0xFF
            value = val & s
            self.write_value(addressname, (value, ))
            return value
            
    def get_encoder_pos(self, motor_number):
        addressname = 'encoder_' +str(motor_number) +'_pos'
        return self.read_value(addressname)[0]
        
    def get_motor_status(self, motor_number, bit_num):
        addressname = 'status_m' + str(motor_number)
        s = self.read_value(addressname)[0]
        x = 1
        x = x << bit_num
        value = x & s
        value = value >> bit_num
        return value

    def get_tasks(self, motor_number):
        addressname = 'tasks_running_m' + str(motor_number)
        return self.read_value(addressname)[0]

    def set_pid(self, pid, target, value):
        addressname = str(pid) + '_' + str(target)
        self.write_value(addressname, (value, ))

    def set_pass_count(self, value):
        self.write_value('pass_count', (value, ))
        
    def set_tolerance(self, value):
        self.write_value('tolerance', (value, ))
        
MMX.add_compatible_sensor(None, 'mndsnsrs', 'NxTMMX') #Tested with version 'V1.01'


class HID(BaseDigitalSensor):
    """Class for Human Interface Device sensors.
These are connected to a computer and look like a keyboard to it."""
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'command'  :        (0x41, '<B'),
                        'modifier'  :       (0x42, '<B'),
                        'keyboard_data'  :  (0x43, '<B'),
                        })
    I2C_DEV = 0x04
    
    class Commands:
        TRANSMIT = 'T'
        ASCII_MODE = 'A'
        DIRECT_MODE = 'D'
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)
    
    def command(self, command):
        value = ord(command)
        self.write_value('command', (value, ))
    
    def set_modifier(self, mod):
        self.write_value('modifier', (mod, ))
    
    def write_data(self, data):
        data = ord(data)
        self.write_value('keyboard_data', (data, ))
                
HID.add_compatible_sensor(None, 'mndsnsrs', 'NXTHID') #Tested with version 'V1.02'


class PS2(BaseDigitalSensor):   
    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update({'command'  :        (0x41, '<B'),
                        'button_set_1':     (0x42, '<B'),
                        'button_set_2':     (0x43, '<B'),
                        'x_left_joystick':  (0x44, '<b'),
                        'y_left_joystick':  (0x45, '<b'),
                        'x_right_joystick': (0x46, '<b'),
                        'y_right_joystick': (0x47, '<b'),
                        })
    
    class ControllerState:
        class Buttons:
            left, down, right, up, square, cross, circle, triangle, r1, r2, r3, l1, l2, l3 = [0 for i in range(14)] #14 zeros
        def __init__(self, buttons_1, buttons_2, left_x, left_y, right_x, right_y):
            self.leftstick = (left_x, left_y)
            self.rightstick = (right_x, right_y)
            buttons_1 = ~buttons_1
            buttons_2 = ~buttons_2
            self.buttons = self.Buttons()
            self.buttons.left = bool(buttons_1 & 0x80)
            self.buttons.down = bool(buttons_1 & 0x40)
            self.buttons.right = bool(buttons_1 & 0x20)
            self.buttons.up = bool(buttons_1 & 0x10)
            self.buttons.square = bool(buttons_2 & 0x80)
            self.buttons.cross = bool(buttons_2 & 0x40)
            self.buttons.circle = bool(buttons_2 & 0x20)
            self.buttons.triangle = bool(buttons_2 & 0x10)
            self.buttons.r1 = bool(buttons_2 & 0x08)
            self.buttons.r2 = bool(buttons_2 & 0x02)
            self.buttons.r3 = bool(buttons_1 & 0x04)
            self.buttons.l1 = bool(buttons_2 & 0x04)
            self.buttons.l2 = bool(buttons_2 & 0x01)
            self.buttons.l3 = bool(buttons_1 & 0x02)
    
    class Commands:
        POWER_ON = 'E'
        POWER_OFF = 'D'
        DIGITAL_MODE = 'A'
        ANALOG_MODE = 's'
        ADPA_ON = 'N'
        ADPA_OFF = 'O'
    
    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)
    
    def command(self, command):
        value = ord(command)
        self.write_value('command', (value, ))

    def get_joystick(self, xy, lr):
        addressname = str(xy) + '_' + str(lr) + '_joystick'
        return self.read_value(addressname)[0]

    def get_buttons(self, setnum):
        addressname = 'button_set_' + str(setnum)
        return self.read_value(addressname)[0]
    
    def get_sample(self):
        return self.ControllerState(
            self.get_buttons(1),
            self.get_buttons(2),
            self.get_joystick('x', 'left'),
            self.get_joystick('y', 'left'),
            self.get_joystick('x', 'right'),
            self.get_joystick('y', 'right'))

PS2.add_compatible_sensor(None, 'mndsnsrs', 'PSPNX') #Tested with version 'V2.00'



# test_sensors -- Test nxt.sensor modules
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
import struct
from unittest.mock import Mock, call

import pytest

import nxt.sensor
import nxt.sensor.analog
import nxt.sensor.digital
import nxt.sensor.generic
import nxt.sensor.hitechnic
import nxt.sensor.mindsensors
from nxt.sensor import Mode, Port, Type


@pytest.fixture
def mdigital(monkeypatch):
    m = Mock(spec_set=("read_value", "write_value"))
    monkeypatch.setattr(
        nxt.sensor.digital.BaseDigitalSensor, "read_value", m.read_value
    )
    monkeypatch.setattr(
        nxt.sensor.digital.BaseDigitalSensor, "write_value", m.write_value
    )
    return m


class TestGeneric:
    """Test non digital sensors."""

    def test_analog(self, mbrick):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.analog.BaseAnalogSensor)
        mbrick.get_input_values.side_effect = [
            (Port.S1, True, False, Type.SWITCH, Mode.BOOL, 1, 2, 3, 4),
        ]
        s.set_input_mode(Type.SWITCH, Mode.BOOL)
        v = s.get_input_values()
        assert v.port == Port.S1
        assert v.valid is True
        assert v.calibrated is False
        assert v.sensor_type == Type.SWITCH
        assert v.mode == Mode.BOOL
        assert v.raw_value == 1
        assert v.normalized_value == 2
        assert v.scaled_value == 3
        assert v.calibrated_value == 4
        assert str(v).startswith("(")
        s.reset_input_scaled_value()
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.SWITCH, Mode.BOOL),
            call.get_input_values(Port.S1),
            call.reset_input_scaled_value(Port.S1),
        ]

    def test_touch(self, mbrick):
        assert (
            nxt.sensor.generic.Touch.get_sample is nxt.sensor.generic.Touch.is_pressed
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.generic.Touch)
        mbrick.get_input_values.side_effect = [
            (Port.S1, True, False, Type.SWITCH, Mode.BOOL, 1023, 1023, 0, 1023),
            (Port.S1, True, False, Type.SWITCH, Mode.BOOL, 183, 183, 1, 183),
        ]
        assert s.is_pressed() is False
        assert s.is_pressed() is True
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.SWITCH, Mode.BOOL),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
        ]

    def test_light(self, mbrick):
        assert (
            nxt.sensor.generic.Light.get_sample
            is nxt.sensor.generic.Light.get_lightness
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.generic.Light)
        mbrick.get_input_values.side_effect = [
            (Port.S1, True, False, Type.LIGHT_ACTIVE, Mode.RAW, 726, 250, 250, 250),
            (Port.S1, True, False, Type.LIGHT_INACTIVE, Mode.RAW, 823, 107, 107, 107),
        ]
        assert s.get_lightness() == 250
        s.set_illuminated(False)
        assert s.get_lightness() == 107
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.LIGHT_ACTIVE, Mode.RAW),
            call.get_input_values(Port.S1),
            call.set_input_mode(Port.S1, Type.LIGHT_INACTIVE, Mode.RAW),
            call.get_input_values(Port.S1),
        ]

    def test_sound(self, mbrick):
        assert (
            nxt.sensor.generic.Sound.get_sample is nxt.sensor.generic.Sound.get_loudness
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.generic.Sound)
        mbrick.get_input_values.side_effect = [
            (Port.S1, True, False, Type.SOUND_DBA, Mode.RAW, 999, 15, 15, 15),
            (Port.S1, True, False, Type.SOUND_DB, Mode.RAW, 999, 15, 15, 15),
        ]
        assert s.get_loudness() == 15
        s.set_adjusted(False)
        assert s.get_loudness() == 15
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.SOUND_DBA, Mode.RAW),
            call.get_input_values(Port.S1),
            call.set_input_mode(Port.S1, Type.SOUND_DB, Mode.RAW),
            call.get_input_values(Port.S1),
        ]

    def test_color(self, mbrick):
        assert nxt.sensor.generic.Color.get_sample is nxt.sensor.generic.Color.get_color
        s = mbrick.get_sensor(Port.S1, nxt.sensor.generic.Color)
        mbrick.get_input_values.side_effect = [
            (Port.S1, True, False, Type.COLOR_FULL, Mode.RAW, 0, 0, 4, 0),
            (Port.S1, True, False, Type.COLOR_FULL, Mode.RAW, 0, 0, 4, 0),
            # TODO: handle invalid measures on configuration change.
            (Port.S1, True, False, Type.COLOR_RED, Mode.RAW, 114, 46, 46, 46),
            (Port.S1, True, False, Type.COLOR_RED, Mode.RAW, 114, 46, 46, 46),
        ]
        color = s.get_color()
        assert color == 4
        assert color == s.DetectedColor.YELLOW
        assert s.get_reflected_light(Type.COLOR_RED) == 46
        assert s.get_light_color() == Type.COLOR_RED
        assert mbrick.mock_calls == [
            # TODO: set too much input mode.
            call.set_input_mode(Port.S1, Type.COLOR_FULL, Mode.RAW),
            call.set_input_mode(Port.S1, Type.COLOR_FULL, Mode.RAW),
            # TODO: get too much input values.
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
            call.set_input_mode(Port.S1, Type.COLOR_RED, Mode.RAW),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
        ]


@pytest.mark.usefixtures("mtime")
class TestDigital:
    """Test nxt.sensor.digital."""

    version_bin = b"V1.0\0\0\0\0"
    product_id_bin = b"LEGO\0\0\0\0"
    sensor_type_bin = b"Sonar\0\0\0"

    def test_get_sensor_info(self, mbrick):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.digital.BaseDigitalSensor, False)
        mbrick.ls_get_status.return_value = 8
        mbrick.ls_read.side_effect = [
            self.version_bin,
            self.product_id_bin,
            self.sensor_type_bin,
        ]
        info = s.get_sensor_info()
        assert info.version == "V1.0"
        assert info.product_id == "LEGO"
        assert info.sensor_type == "Sonar"
        print(info)
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.LOW_SPEED_9V, Mode.RAW),
            call.ls_write(Port.S1, bytes((0x02, 0x00)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
            call.ls_write(Port.S1, bytes((0x02, 0x08)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
            call.ls_write(Port.S1, bytes((0x02, 0x10)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
        ]

    def test_check_compatible(self, mbrick, caplog):
        class DummySensor(nxt.sensor.digital.BaseDigitalSensor):
            pass

        DummySensor.add_compatible_sensor(None, "NXT-PYTH", "Dummy")
        mbrick.ls_get_status.return_value = 8
        mbrick.ls_read.side_effect = [
            b"V3\0\0\0\0\0\0",
            b"NXT-PYTH",
            b"Dummy\0\0\0",
            b"V2\0\0\0\0\0\0",
            b"NXT-PYTH",
            b"Plop\0\0\0\0",
        ]
        DummySensor(mbrick, Port.S1)
        assert "WARNING" not in caplog.text
        DummySensor(mbrick, Port.S2)
        assert "WARNING" in caplog.text

    def test_write_value(self, mbrick):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.digital.BaseDigitalSensor, False)
        s.I2C_ADDRESS = dict(s.I2C_ADDRESS, command=(0x41, "B"))
        s.write_value("command", (0x12,))
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.LOW_SPEED_9V, Mode.RAW),
            call.ls_write(Port.S1, bytes((0x02, 0x41, 0x12)), 0),
        ]

    def test_not_ready(self, mbrick):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.digital.BaseDigitalSensor, False)
        mbrick.ls_get_status.side_effect = [nxt.error.I2CPendingError("pending"), 8]
        mbrick.ls_read.return_value = self.product_id_bin
        assert s.read_value("product_id") == (self.product_id_bin,)
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.LOW_SPEED_9V, Mode.RAW),
            call.ls_write(Port.S1, bytes((0x02, 0x08)), 8),
            call.ls_get_status(Port.S1),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
        ]

    def test_status_timeout(self, mbrick):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.digital.BaseDigitalSensor, False)
        mbrick.ls_get_status.side_effect = (
            [nxt.error.I2CPendingError("pending")] * 30 * 3
        )
        with pytest.raises(nxt.error.I2CError):
            s.read_value("product_id")
        mock_calls = [call.set_input_mode(Port.S1, Type.LOW_SPEED_9V, Mode.RAW)] + (
            [call.ls_write(Port.S1, bytes((0x02, 0x08)), 8)]
            + [call.ls_get_status(Port.S1)] * 30
            + [call.ls_read(Port.S1)]
        ) * 3
        assert mbrick.mock_calls == mock_calls

    def test_read_error(self, mbrick):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.digital.BaseDigitalSensor, False)
        mbrick.ls_get_status.side_effect = [8, 8]
        mbrick.ls_read.side_effect = [self.product_id_bin[1:], self.product_id_bin]
        assert s.read_value("product_id") == (self.product_id_bin,)
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.LOW_SPEED_9V, Mode.RAW),
            call.ls_write(Port.S1, bytes((0x02, 0x08)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
            # Retry.
            call.ls_write(Port.S1, bytes((0x02, 0x08)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
        ]

    def test_read_timeout(self, mbrick):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.digital.BaseDigitalSensor, False)
        mbrick.ls_get_status.return_value = 8
        mbrick.ls_read.return_value = self.product_id_bin[1:]
        with pytest.raises(nxt.error.I2CError):
            s.read_value("product_id")
        mock_calls = [call.set_input_mode(Port.S1, Type.LOW_SPEED_9V, Mode.RAW)] + [
            call.ls_write(Port.S1, bytes((0x02, 0x08)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
        ] * 3
        assert mbrick.mock_calls == mock_calls

    def test_find_class(self):
        def test(info, cls):
            found = nxt.sensor.digital.find_class(nxt.sensor.digital.SensorInfo(*info))
            assert found is cls

        test(("V1.0", "LEGO", "Sonar"), nxt.sensor.generic.Ultrasonic)
        test(("Vx.xx", "mndsnsrs", "CMPS"), nxt.sensor.mindsensors.Compassv2)
        test(("Vx.xx", "mndsnsrs", "DIST"), nxt.sensor.mindsensors.DIST)
        test(("V3.20", "mndsnsrs", "ACCL-NX"), nxt.sensor.mindsensors.ACCL)
        test(("V2.11", "mndsnsrs", "MTRMUX"), nxt.sensor.mindsensors.MTRMUX)
        test(("V1.16", "mndsnsrs", "LineLdr"), nxt.sensor.mindsensors.LineLeader)
        test(("V1.20", "mndsnsrs", "NXTServo"), nxt.sensor.mindsensors.Servo)
        test(("V1.01", "mndsnsrs", "NxTMMX"), nxt.sensor.mindsensors.MMX)
        test(("V1.02", "mndsnsrs", "NXTHID"), nxt.sensor.mindsensors.HID)
        test(("V2.00", "mndsnsrs", "PSPNX"), nxt.sensor.mindsensors.PS2)
        test(("\xfdV1.23", "HiTechnc", "Compass "), nxt.sensor.hitechnic.Compass)
        test(("\xfdV2.1", "HITECHNC", "Compass "), nxt.sensor.hitechnic.Compass)
        test(
            ("\xfdV1.1   ", "HITECHNC", "Accel.  "), nxt.sensor.hitechnic.Accelerometer
        )
        test(("Vx.xx", "HITECHNC", "IRRecv  "), nxt.sensor.hitechnic.IRReceiver)
        test(("Vx.xx", "HITECHNC", "NewIRDir"), nxt.sensor.hitechnic.IRSeekerv2)
        test(("Vx.xx", "HITECHNC", "ColorPD"), nxt.sensor.hitechnic.Colorv2)
        test(("Vx.xx", "HITECHNC", "ColorPD "), nxt.sensor.hitechnic.Colorv2)
        test(("Vx.xx", "HiTechnc", "Proto   "), nxt.sensor.hitechnic.Prototype)
        test(("Vx.xx", "HiTechnc", "ServoCon"), nxt.sensor.hitechnic.ServoCon)
        test(("Vx.xx", "HiTechnc", "MotorCon"), nxt.sensor.hitechnic.MotorCon)

    def test_get_sensor(self, mbrick):
        mbrick.ls_get_status.return_value = 8
        mbrick.ls_read.side_effect = [
            self.version_bin,
            self.product_id_bin,
            self.sensor_type_bin,
        ]
        assert mbrick.get_sensor(Port.S1).__class__ is nxt.sensor.generic.Ultrasonic
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.LOW_SPEED_9V, Mode.RAW),
            call.ls_write(Port.S1, bytes((0x02, 0x00)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
            call.ls_write(Port.S1, bytes((0x02, 0x08)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
            call.ls_write(Port.S1, bytes((0x02, 0x10)), 8),
            call.ls_get_status(Port.S1),
            call.ls_read(Port.S1),
            call.set_input_mode(Port.S1, Type.LOW_SPEED_9V, Mode.RAW),
        ]


class TestGenericDigital:
    """Test LEGO digital sensors."""

    def test_ultrasonic(self, mbrick, mdigital):
        assert (
            nxt.sensor.generic.Ultrasonic.get_sample
            is nxt.sensor.generic.Ultrasonic.get_distance
        )
        s = mbrick.get_sensor(
            Port.S1, nxt.sensor.generic.Ultrasonic, check_compatible=False
        )
        mdigital.read_value.side_effect = [
            (42,),
            (b"10E-2m\0",),
            (1, 2, 3, 4, 5, 6, 7, 8),
            (4,),
            (43,),
        ]
        assert s.get_distance() == 42
        assert s.get_measurement_units() == "10E-2m"
        assert s.get_all_measurements() == (1, 2, 3, 4, 5, 6, 7, 8)
        assert s.get_measurement_no(3) == 4
        s.command(s.Command.OFF)
        assert s.get_interval() == 43
        s.set_interval(44)
        assert mdigital.mock_calls == [
            call.read_value("measurement_byte_0"),
            call.read_value("measurement_units"),
            call.read_value("measurements"),
            call.read_value("measurement_byte_3"),
            call.write_value("command", (0,)),
            call.read_value("continuous_measurement_interval"),
            call.write_value("continuous_measurement_interval", (44,)),
        ]

    def test_temperature(self, mbrick, mdigital):
        assert (
            nxt.sensor.generic.Temperature.get_sample
            is nxt.sensor.generic.Temperature.get_deg_c
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.generic.Temperature)
        mdigital.read_value.side_effect = [
            (1600 * 16,),
            (1600 * 16,),
            struct.unpack(">h", b"\xff\x80"),
        ]
        assert s.get_deg_c() == 100
        assert s.get_deg_f() == 212
        assert s.get_deg_c() == -0.5
        assert mdigital.mock_calls == [
            call.read_value("raw_value"),
            call.read_value("raw_value"),
            call.read_value("raw_value"),
        ]


class TestMindsensors:
    """Test Mindsensors sensors."""

    def test_sumoeyes(self, mbrick):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.SumoEyes)
        mbrick.get_input_values.side_effect = [
            (Port.S1, True, False, Type.LIGHT_ACTIVE, Mode.RAW, 0, 0, 0, 0),
            (Port.S1, True, False, Type.LIGHT_ACTIVE, Mode.RAW, 0, 350, 0, 0),
            (Port.S1, True, False, Type.LIGHT_ACTIVE, Mode.RAW, 0, 650, 0, 0),
            (Port.S1, True, False, Type.LIGHT_ACTIVE, Mode.RAW, 0, 800, 0, 0),
        ]
        m = s.get_sample()
        assert str(m) == "(left: False, right: False)"
        assert (m.left, m.right) == (False, False)
        m = s.get_sample()
        assert (m.left, m.right) == (True, False)
        m = s.get_sample()
        assert (m.left, m.right) == (False, True)
        m = s.get_sample()
        assert (m.left, m.right) == (True, True)
        s.set_long_range(True)
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.LIGHT_ACTIVE, Mode.RAW),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
            call.set_input_mode(Port.S1, Type.LIGHT_INACTIVE, Mode.RAW),
        ]

    def test_compassv2(self, mbrick, mdigital):
        assert (
            nxt.sensor.mindsensors.Compassv2.get_sample
            is nxt.sensor.mindsensors.Compassv2.get_heading
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.Compassv2, False)
        mdigital.read_value.return_value = (300,)
        # TODO: should return degrees (divide by 10).
        assert s.get_heading() == 300
        s.command(s.Commands.BEGIN_CALIBRATION)
        # TODO: get rid of ord, adapt format.
        assert mdigital.mock_calls == [
            call.write_value("command", (ord("I"),)),
            call.read_value("heading"),
            call.write_value("command", (ord("C"),)),
        ]

    def test_dist(self, mbrick, mdigital):
        assert (
            nxt.sensor.mindsensors.DIST.get_sample
            is nxt.sensor.mindsensors.DIST.get_distance
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.DIST, False)
        # TODO: get rid of ord, adapt format.
        mdigital.read_value.side_effect = [(100,), (ord("2"),), (42,), (43,), (44,)]
        assert s.get_distance() == 100
        assert s.get_type() == ord(s.Commands.TYPE_GP2D120)
        s.command(s.Commands.POWER_OFF)
        assert s.get_voltage() == 42
        assert s.get_min_distance() == 43
        assert s.get_max_distance() == 44
        assert mdigital.mock_calls == [
            call.read_value("distance"),
            call.read_value("type"),
            call.write_value("command", (ord("D"),)),
            call.read_value("voltage"),
            call.read_value("min_distance"),
            call.read_value("max_distance"),
        ]

    def test_rtc(self, mbrick, mdigital):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.RTC)
        # TODO: this one is completely broken:
        #  - Return str instead of int.
        #  - Bad handling of hour format.
        #  - Fields are read one by one (what happen if you read minutes at 11:59 and
        #    hours at 12:00?
        #  - Pretty sure struct_time is not right too.
        mdigital.read_value.side_effect = [(0x32,)]
        assert s.get_seconds() == "32"

    def test_accl(self, mbrick, mdigital):
        assert (
            nxt.sensor.mindsensors.ACCL.get_sample
            is nxt.sensor.mindsensors.ACCL.get_all_accel
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.ACCL, False)
        # TODO: get rid of ord, adapt format.
        mdigital.read_value.side_effect = [
            (ord("2"),),
            (42,),
            (1, 2, 3),
            (43,),
            (1, 2, 3),
            (44,),
            (45,),
        ]
        s.command(s.Commands.SENS_2G)
        assert s.get_sensitivity() == s.Commands.SENS_2G
        assert s.get_tilt("x") == 42
        assert s.get_all_tilt() == (1, 2, 3)
        assert s.get_accel("z") == 43
        assert s.get_all_accel() == (1, 2, 3)
        assert s.get_offset("x") == 44
        assert s.get_range("x") == 45
        s.set_offset("x", 46)
        s.set_range("x", 47)
        assert mdigital.mock_calls == [
            call.write_value("command", (ord("2"),)),
            call.read_value("sensitivity"),
            call.read_value("x_tilt"),
            call.read_value("all_tilt"),
            call.read_value("z_accel"),
            call.read_value("all_accel"),
            call.read_value("x_offset"),
            call.read_value("x_range"),
            call.write_value("x_offset", (46,)),
            call.write_value("x_range", (47,)),
        ]

    def test_mtrmux(self, mbrick, mdigital):
        assert not hasattr(nxt.sensor.mindsensors.MTRMUX, "get_sample")
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.MTRMUX, False)
        mdigital.read_value.side_effect = [(1,), (2,)]
        s.command(s.Commands.FLOAT)
        s.set_direction(1, 1)
        s.set_speed(2, 2)
        assert s.get_direction(3) == 1
        assert s.get_speed(4) == 2
        assert mdigital.mock_calls == [
            call.write_value("command", (0,)),
            call.write_value("direction_m1", (1,)),
            call.write_value("speed_m2", (2,)),
            call.read_value("direction_m3"),
            call.read_value("speed_m4"),
        ]

    def test_lineleader(self, mbrick, mdigital):
        assert (
            nxt.sensor.mindsensors.LineLeader.get_sample
            is nxt.sensor.mindsensors.LineLeader.get_reading_all
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.LineLeader, False)
        mdigital.read_value.side_effect = [
            (-10,),
            (50,),
            (0x5A,),
            (50,),
            (1, 2, 3, 4, 5, 6, 7, 8),
            (60,),
            (11, 12, 13, 14, 15, 16, 17, 18),
        ]
        s.command(s.Commands.CALIBRATE_WHITE)
        assert s.get_steering() == -10
        assert s.get_average() == 50
        assert s.get_result() == 0x5A
        s.set_set_point(50)
        s.set_pid("p", 35)
        s.set_pid_divisor("i", 10)
        assert s.get_reading(1) == 50
        assert s.get_reading_all() == (1, 2, 3, 4, 5, 6, 7, 8)
        assert s.get_uncal_reading(2) == 60
        assert s.get_uncal_all() == (11, 12, 13, 14, 15, 16, 17, 18)
        assert mdigital.mock_calls == [
            # TODO: get rid of ord, adapt format.
            call.write_value("command", (ord("W"),)),
            call.read_value("steering"),
            call.read_value("average"),
            call.read_value("result"),
            call.write_value("set_point", (50,)),
            call.write_value("kp", (35,)),
            call.write_value("ki_divisor", (10,)),
            call.read_value("calibrated_reading_byte1"),
            call.read_value("all_calibrated_readings"),
            call.read_value("uncal_sensor2_voltage_byte1"),
            call.read_value("all_uncal_readings"),
        ]

    def test_servo(self, mbrick, mdigital):
        assert not hasattr(nxt.sensor.mindsensors.Servo, "get_sample")
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.Servo, False)
        # TODO: command can not work, can not fit two bytes in one byte.
        mdigital.read_value.side_effect = [(1,), (42,), (43,)]
        assert s.get_bat_level() == 1
        s.set_position(1, 42)
        assert s.get_position(1) == 42
        s.set_speed(2, 43)
        assert s.get_speed(2) == 43
        s.set_quick(3, 44)
        assert mdigital.mock_calls == [
            call.read_value("command"),
            call.write_value("servo_1_pos", (42,)),
            call.read_value("servo_1_pos"),
            call.write_value("servo_2_speed", (43,)),
            call.read_value("servo_2_speed"),
            call.write_value("servo_3_quick", (44,)),
        ]

    def test_mmx(self, mbrick, mdigital):
        assert not hasattr(nxt.sensor.mindsensors.MMX, "get_sample")
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.MMX, False)
        mdigital.read_value.side_effect = [
            (1,),
            (0xAA,),
            (0x55,),
            (12345,),
            (0x55,),
            (42,),
        ]
        # TODO: get rid of ord, adapt format.
        s.command(s.Commands.RESET_PARAMS_ENCODERS)
        assert s.get_bat_level() == 1
        s.set_encoder_target(1, 12345)
        s.set_speed(2, 50)
        s.set_time_run(1, 60)
        s.command_b(1, 42)
        s.command_a(2, 2, 1)
        s.command_a(2, 2, 0)
        assert s.get_encoder_pos(1) == 12345
        # TODO: should be bool.
        assert s.get_motor_status(1, 2) == 1
        assert s.get_tasks(1) == 42
        s.set_pid("p", "speed", 3)
        s.set_pass_count(43)
        s.set_tolerance(44)
        assert mdigital.mock_calls == [
            call.write_value("command", (ord("R"),)),
            call.read_value("command"),
            call.write_value("encoder_1_target", (12345,)),
            call.write_value("speed_2", (50,)),
            call.write_value("seconds_to_run_1", (60,)),
            call.write_value("command_b_1", (42,)),
            call.read_value("command_a_2"),
            call.write_value("command_a_2", (0xAE,)),
            call.read_value("command_a_2"),
            call.write_value("command_a_2", (0x51,)),
            call.read_value("encoder_1_pos"),
            call.read_value("status_m1"),
            call.read_value("tasks_running_m1"),
            call.write_value("p_speed", (3,)),
            call.write_value("pass_count", (43,)),
            call.write_value("tolerance", (44,)),
        ]

    def test_hid(self, mbrick, mdigital):
        assert not hasattr(nxt.sensor.mindsensors.HID, "get_sample")
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.HID, False)
        s.command(s.Commands.ASCII_MODE)
        s.set_modifier(42)
        s.write_data("a")
        assert mdigital.mock_calls == [
            call.write_value("command", (ord("A"),)),
            call.write_value("modifier", (42,)),
            call.write_value("keyboard_data", (ord("a"),)),
        ]

    def test_ps2(self, mbrick, mdigital):
        s = mbrick.get_sensor(Port.S1, nxt.sensor.mindsensors.PS2, False)
        mdigital.read_value.side_effect = [
            (42,),
            (0x55,),
            # TODO: read everything in one read.
            (0x55,),
            (0x55,),
            (42,),
            (43,),
            (44,),
            (45,),
        ]
        s.command(s.Commands.POWER_ON)
        assert s.get_joystick("x", "left") == 42
        assert s.get_buttons(1) == 0x55
        st = s.get_sample()
        assert st.leftstick == (42, 43)
        assert st.rightstick == (44, 45)
        assert st.buttons.left is True
        assert st.buttons.down is False
        assert st.buttons.right is True
        assert st.buttons.up is False
        assert st.buttons.square is True
        assert st.buttons.cross is False
        assert st.buttons.circle is True
        assert st.buttons.triangle is False
        assert st.buttons.r1 is True
        assert st.buttons.r2 is True
        assert st.buttons.r3 is False
        assert st.buttons.l1 is False
        assert st.buttons.l2 is False
        assert st.buttons.l3 is True
        assert mdigital.mock_calls == [
            call.write_value("command", (ord("E"),)),
            call.read_value("x_left_joystick"),
            call.read_value("button_set_1"),
            call.read_value("button_set_1"),
            call.read_value("button_set_2"),
            call.read_value("x_left_joystick"),
            call.read_value("y_left_joystick"),
            call.read_value("x_right_joystick"),
            call.read_value("y_right_joystick"),
        ]


class TestHitechnic:
    """Test HiTechnic sensors."""

    def test_compass(self, mbrick, mdigital):
        assert (
            nxt.sensor.hitechnic.Compass.get_sample
            is nxt.sensor.hitechnic.Compass.get_heading
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.Compass, False)
        mdigital.read_value.return_value = (10,)
        assert s.get_heading() == 30
        assert s.get_relative_heading(0) == 30
        assert s.get_relative_heading(-170) == -160
        mdigital.read_value.return_value = (-10,)
        assert s.get_relative_heading(170) == 160
        assert s.is_in_range(-40, -20) is True
        assert s.is_in_range(-20, -40) is False
        mdigital.read_value.return_value = (0x02,)
        assert s.get_mode() == s.Modes.CALIBRATION_FAILED
        s.set_mode(s.Modes.CALIBRATION)
        with pytest.raises(ValueError):
            s.set_mode(s.Modes.CALIBRATION_FAILED)
        assert mdigital.mock_calls == [
            call.read_value("heading"),
            call.read_value("adder"),
        ] * 6 + [
            call.read_value("mode"),
            call.write_value("mode", (0x43,)),
        ]

    def test_accelerometer(self, mbrick, mdigital):
        assert (
            nxt.sensor.hitechnic.Accelerometer.get_sample
            is nxt.sensor.hitechnic.Accelerometer.get_acceleration
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.Accelerometer, False)
        mdigital.read_value.return_value = (0x12, 0x23, -0x32, 0x3, 0x0, 0x2)
        v = s.get_acceleration()
        assert (v.x, v.y, v.z) == (75, 140, -198)
        assert mdigital.mock_calls == [
            call.read_value("all_data"),
        ]

    def test_irreceiver(self, mbrick, mdigital):
        assert (
            nxt.sensor.hitechnic.IRReceiver.get_sample
            is nxt.sensor.hitechnic.IRReceiver.get_speeds
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.IRReceiver, False)
        mdigital.read_value.return_value = (0, -16, 30, -44, 58, -72, 100, -128)
        v = s.get_speeds()
        assert (v.m1A, v.m1B) == (0, -16)
        assert v.channel_1 == (0, -16)
        assert (v.m2A, v.m2B) == (30, -44)
        assert v.channel_2 == (30, -44)
        assert (v.m3A, v.m3B) == (58, -72)
        assert v.channel_3 == (58, -72)
        # TODO: handle -128 specially.
        assert (v.m4A, v.m4B) == (100, -128)
        assert v.channel_4 == (100, -128)
        assert mdigital.mock_calls == [
            call.read_value("all_data"),
        ]

    def test_irseekerv2(self, mbrick, mdigital):
        assert (
            nxt.sensor.hitechnic.IRSeekerv2.get_sample
            is nxt.sensor.hitechnic.IRSeekerv2.get_ac_values
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.IRSeekerv2, False)
        mdigital.read_value.side_effect = [
            (5, 42, 43, 44, 45, 46, 44),
            (5, 42, 43, 44, 45, 46),
            (0,),
        ]
        v = s.get_dc_values()
        assert v.direction == 5
        # TODO: use a tuple.
        assert v.sensor_1 == 42
        assert v.sensor_2 == 43
        assert v.sensor_3 == 44
        assert v.sensor_4 == 45
        assert v.sensor_5 == 46
        assert v.sensor_mean == 44
        assert v.get_dir_brightness(5) == 44
        assert v.get_dir_brightness(4) == 43.5
        v = s.get_ac_values()
        assert v.direction == 5
        assert v.sensor_1 == 42
        assert v.sensor_2 == 43
        assert v.sensor_3 == 44
        assert v.sensor_4 == 45
        assert v.sensor_5 == 46
        assert v.get_dir_brightness(5) == 44
        assert v.get_dir_brightness(4) == 43.5
        assert s.get_dsp_mode() == s.DSPModes.AC_DSP_1200Hz
        s.set_dsp_mode(s.DSPModes.AC_DSP_600Hz)
        assert mdigital.mock_calls == [
            call.read_value("all_DC"),
            call.read_value("all_AC"),
            call.read_value("dspmode"),
            call.write_value("dspmode", (1,)),
        ]

    def test_eopd(self, mbrick):
        assert (
            nxt.sensor.hitechnic.EOPD.get_sample
            is nxt.sensor.hitechnic.EOPD.get_scaled_value
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.EOPD)
        mbrick.get_input_values.side_effect = [
            (Port.S1, True, False, Type.LIGHT_INACTIVE, Mode.RAW, 523, 0, 0, 0),
            (Port.S1, True, False, Type.LIGHT_INACTIVE, Mode.RAW, 398, 0, 0, 0),
            (Port.S1, True, False, Type.LIGHT_INACTIVE, Mode.RAW, 398, 0, 0, 0),
            (Port.S1, True, False, Type.LIGHT_INACTIVE, Mode.RAW, 1023, 0, 0, 0),
        ]
        # TODO: choose a mode in constructor, or raise error if no mode chosen.
        s.set_range_long()
        s.set_range_short()
        assert s.get_raw_value() == 500
        assert s.get_processed_value() == 25
        assert s.get_scaled_value() == 10
        assert s.get_scaled_value() == 250
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.LIGHT_ACTIVE, Mode.RAW),
            call.set_input_mode(Port.S1, Type.LIGHT_INACTIVE, Mode.RAW),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
        ]

    def test_colorv2(self, mbrick, mdigital):
        assert (
            nxt.sensor.hitechnic.Colorv2.get_sample
            is nxt.sensor.hitechnic.Colorv2.get_active_color
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.Colorv2, False)
        mdigital.read_value.side_effect = [
            (8, 100, 50, 0, 75, 42, 66, 33, 0),
            (100, 50, 0, 75),
            (0,),
        ]
        v = s.get_active_color()
        assert v.number == 8
        assert v.red == 100
        assert v.green == 50
        assert v.blue == 0
        assert v.white == 75
        assert v.index == 42
        assert v.normred == 66
        assert v.normgreen == 33
        assert v.normblue == 0
        v = s.get_passive_color()
        assert v.red == 100
        assert v.green == 50
        assert v.blue == 0
        assert v.white == 75
        assert s.get_mode() == s.Modes.ACTIVE
        s.set_mode(s.Modes.PASSIVE)
        assert mdigital.mock_calls == [
            call.read_value("all_data"),
            call.read_value("all_raw_data"),
            call.read_value("mode"),
            call.write_value("mode", (1,)),
        ]

    def test_giro(self, mbrick):
        assert (
            nxt.sensor.hitechnic.Gyro.get_sample
            is nxt.sensor.hitechnic.Gyro.get_rotation_speed
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.Gyro)
        mbrick.get_input_values.side_effect = [
            (Port.S1, True, False, Type.ANGLE, Mode.RAW, 0, 0, 42, 0),
            (Port.S1, True, False, Type.ANGLE, Mode.RAW, 0, 0, 42, 0),
            (Port.S1, True, False, Type.ANGLE, Mode.RAW, 0, 0, 54, 0),
            (Port.S1, True, False, Type.ANGLE, Mode.RAW, 0, 0, 54, 0),
        ]
        assert s.get_rotation_speed() == 42
        s.calibrate()
        assert s.get_rotation_speed() == 12
        s.set_zero(0)
        assert s.get_rotation_speed() == 54
        assert mbrick.mock_calls == [
            call.set_input_mode(Port.S1, Type.ANGLE, Mode.RAW),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
            call.get_input_values(Port.S1),
        ]

    def test_prototype(self, mbrick, mdigital):
        assert not hasattr(nxt.sensor.hitechnic.Prototype, "get_sample")
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.Prototype, False)
        mdigital.read_value.side_effect = [
            (42, 43, 44, 45, 46),
            (0x2A,),
        ]
        v = s.get_analog()
        assert v.a0 == 42
        assert v.a1 == 43
        assert v.a2 == 44
        assert v.a3 == 45
        assert v.a4 == 46
        v = s.get_digital()
        assert v.dataint == 0x2A
        assert v.datalst == [False, True, False, True, False, True]
        assert list(v) == [False, True, False, True, False, True]
        assert v[0] is False
        assert v.d0 is False
        assert v.d1 is True
        assert v.d2 is False
        assert v.d3 is True
        assert v.d4 is False
        assert v.d5 is True
        s.set_digital(s.Digital_Data(0x15))
        s.set_digital_modes(s.Digital_Data((False, True, False, True, False, True)))
        assert mdigital.mock_calls == [
            call.read_value("all_analog"),
            call.read_value("digital_in"),
            call.write_value("digital_out", (0x15,)),
            call.write_value("digital_cont", (0x2A,)),
        ]

    def test_servocon(self, mbrick, mdigital):
        assert not hasattr(nxt.sensor.hitechnic.ServoCon, "get_sample")
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.ServoCon, False)
        mdigital.read_value.side_effect = [
            (1,),
            (43,),
        ]
        assert s.get_status() == s.Status.RUNNING
        s.set_step_time(5)
        s.set_pos(1, 42)
        assert s.get_pwm() == 43
        s.set_pwm(44)
        assert mdigital.mock_calls == [
            call.read_value("status"),
            call.write_value("steptime", (5,)),
            call.write_value("s1pos", (42,)),
            call.read_value("pwm"),
            call.write_value("pwm", (44,)),
        ]

    def test_motorcon(self, mbrick, mdigital):
        assert not hasattr(nxt.sensor.hitechnic.MotorCon, "get_sample")
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.MotorCon, False)
        mdigital.read_value.side_effect = [
            (123456,),
            (654321,),
            (0x55,),
            (43,),
            (1,),
            (3, 2, 1),
            (0x70, 0x2),
        ]
        s.set_enc_target(1, 123456)
        assert s.get_enc_target(1) == 123456
        assert s.get_enc_current(2) == 654321
        s.set_mode(1, 0x55)
        assert s.get_mode(2) == 0x55
        s.set_power(1, 42)
        assert s.get_power(2) == 43
        s.set_gear_ratio(1, 2)
        assert s.get_gear_ratio(2) == 1
        s.set_pid(1, s.PID_Data(3, 2, 1))
        v = s.get_pid(2)
        assert (v.p, v.i, v.d) == (3, 2, 1)
        assert s.get_battery_voltage() == 450
        assert mdigital.mock_calls == [
            call.write_value("m1enctarget", (123456,)),
            call.read_value("m1enctarget"),
            call.read_value("m2enccurrent"),
            call.write_value("m1mode", (0x55,)),
            call.read_value("m2mode"),
            call.write_value("m1power", (42,)),
            call.read_value("m2power"),
            call.write_value("m1gearratio", (2,)),
            call.read_value("m2gearratio"),
            call.write_value("m1pid", (3, 2, 1)),
            call.read_value("m2pid"),
            call.read_value("batteryvoltage"),
        ]

    def test_angle(self, mbrick, mdigital):
        assert (
            nxt.sensor.hitechnic.Angle.get_sample
            is nxt.sensor.hitechnic.Angle.get_angle
        )
        s = mbrick.get_sensor(Port.S1, nxt.sensor.hitechnic.Angle, False)
        mdigital.read_value.side_effect = [
            (21, 1),
            (123456789,),
            (16,),
        ]
        assert s.get_angle() == 43
        assert s.get_accumulated_angle() == 123456789
        assert s.get_rpm() == 16
        s.calibrate()
        s.reset()
        assert mdigital.mock_calls == [
            call.read_value("angle"),
            call.read_value("angle_acc"),
            call.read_value("rpm"),
            call.write_value("mode", b"C"),
            call.write_value("mode", b"R"),
        ]



#!/usr/bin/env python3
"""NXT-Python example to use sensors."""
import time

import nxt.locator
import nxt.sensor
import nxt.sensor.generic

with nxt.locator.find() as b:
    touch = b.get_sensor(nxt.sensor.Port.S1, nxt.sensor.generic.Touch)
    sound = b.get_sensor(nxt.sensor.Port.S2, nxt.sensor.generic.Sound)
    light = b.get_sensor(nxt.sensor.Port.S3, nxt.sensor.generic.Light, False)
    us = b.get_sensor(nxt.sensor.Port.S4)
    sensors = [touch, sound, light, us]

    while True:
        samples = [s.get_sample() for s in sensors]
        print(samples)
        time.sleep(0.5)



# nxt.error module -- NXT brick error handling
# Copyright (C) 2006, 2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


class ProtocolError(Exception):
    """Raised on error with the NXT brick protocol."""

    pass


class SystemProtocolError(ProtocolError):
    """Raised on error with the NXT brick protocol for a system command."""

    pass


class FileExistsError(SystemProtocolError):
    """Raised when trying to create a file which already exists."""

    pass


class FileNotFoundError(SystemProtocolError):
    """Raised when trying to access a file which does not exists."""

    pass


class ModuleNotFoundError(SystemProtocolError):
    """Raised when trying to access a module which does not exists."""

    pass


class DirectProtocolError(ProtocolError):
    """Raised on error with the NXT brick protocol for a direct command."""

    pass


class I2CError(DirectProtocolError):
    """Raised on I2C error on a sensor port."""

    pass


class I2CPendingError(I2CError):
    """Raised when an I2C transaction on a sensor port is still in progress."""

    pass



#!/usr/bin/env python3
"""NXT-Python example to use Linus Atorf's MotorControl.

You need to have MotorControl22 program installed on the brick, you can find it here:

https://github.com/schodet/MotorControl

Build it and install it with nxc, or upload the precompiled version (MotorControl22.rxe
file).
"""
import time

import nxt.locator
import nxt.motcont
import nxt.motor

with nxt.locator.find() as b:
    mc = nxt.motcont.MotCont(b)

    def wait():
        while not mc.is_ready(nxt.motor.Port.A) or not mc.is_ready(nxt.motor.Port.B):
            time.sleep(0.5)

    mc.start()

    mc.cmd((nxt.motor.Port.A, nxt.motor.Port.B), 50, 360)
    wait()

    mc.cmd((nxt.motor.Port.A, nxt.motor.Port.B), -50, 360)
    wait()

    mc.stop()



# test_backend_socket -- Test nxt.backend.socket module
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
from unittest.mock import Mock, call, patch

import pytest

import nxt.backend.socket


@pytest.fixture
def mdev():
    dev = Mock(
        spec_set=(
            "connect",
            "send",
            "recv",
            "close",
        )
    )
    return dev


@pytest.fixture
def msocket(mdev):
    with patch("nxt.backend.socket.socket") as socket:
        socket.socket.return_value = mdev
        yield socket


def test_socket(msocket, mdev):
    # Instantiate backend.
    backend = nxt.backend.socket.get_backend()
    # Find brick.
    mdev.recv.return_value = b"usb"
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 1
    brick = bricks[0]
    assert brick._sock.type == "ipusb"
    assert msocket.socket.called
    assert mdev.connect.called
    assert mdev.send.call_args == call(b"\x98")
    sock = brick._sock
    # str.
    assert str(sock) == "Socket (localhost:2727)"
    # Send.
    some_bytes = bytes.fromhex("01020304")
    sock.send(some_bytes)
    assert mdev.send.call_args == call(some_bytes)
    # Recv.
    mdev.recv.return_value = some_bytes
    r = sock.recv()
    assert r == some_bytes
    assert mdev.recv.called
    # Close.
    brick.close()
    assert mdev.send.call_args == call(b"\x99")
    assert mdev.close.called
    # Duplicated close.
    sock.close()


def test_socket_cant_connect(msocket, mdev):
    backend = nxt.backend.socket.get_backend()
    mdev.connect.side_effect = [ConnectionRefusedError]
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 0


@pytest.mark.nxt("socket")
def test_socket_real():
    # Instantiate backend.
    backend = nxt.backend.socket.get_backend()
    # Find brick.
    bricks = list(backend.find())
    assert len(bricks) > 0, "no NXT found"
    brick = bricks[0]
    sock = brick._sock
    # str.
    assert str(sock) == "Socket (localhost:2727)"
    # Send.
    sock.send(bytes.fromhex("019b"))
    # Recv.
    r = sock.recv()
    assert r.startswith(bytes.fromhex("029b00"))
    # Close.
    brick.close()



#!/usr/bin/python3
"""NXT-Python tutorial: use ultra-sonic sensor."""
import time

import nxt.locator
import nxt.sensor

# Need to import generic sensors for auto-detection to work.
import nxt.sensor.generic

with nxt.locator.find() as b:
    # Find the sensor connected to port 4.
    mysensor = b.get_sensor(nxt.sensor.Port.S4)
    # Read the sensor in a loop (until interrupted).
    print("Use Ctrl-C to interrupt")
    while True:
        distance_cm = mysensor.get_sample()
        print(distance_cm)
        time.sleep(0.5)



# conftest -- Common test fixtures for doctest
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
import argparse
from unittest.mock import Mock, patch

import pytest


@pytest.fixture(autouse=True)
def mock_backends():
    brick = Mock()
    brick.get_device_info.return_value = ("NXT", None, None, None)
    brick.play_tone.return_value = None
    backend = Mock()
    backend.find.return_value = [brick]
    with patch("nxt.locator._get_default_backends") as m:
        m.return_value = [backend]
        yield m


@pytest.fixture(autouse=True)
def mock_config():
    with patch("nxt.locator._get_config") as m:
        m.return_value = None
        yield m


@pytest.fixture(autouse=True)
def mock_argparse(monkeypatch):
    ap = Mock()
    parser = Mock()
    ns = Mock()
    ns.backends = None
    ns.name = None
    ns.host = None
    ap.return_value = parser
    parser.add_argument.return_value = None
    parser.parse_args.return_value = ns
    monkeypatch.setattr(argparse, "ArgumentParser", ap)



# test_backend_bluetooth -- Test nxt.backend.bluetooth module
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
from unittest.mock import Mock, call, patch

import pytest

import nxt.backend.bluetooth


@pytest.fixture
def msock():
    sock = Mock(
        spec_set=(
            "connect",
            "close",
            "send",
            "recv",
        )
    )
    return sock


@pytest.fixture
def mbluetooth(msock):
    class BluetoothError(Exception):
        pass

    bluetooth = Mock()
    bluetooth.BluetoothSocket.return_value = msock
    bluetooth.BluetoothError = BluetoothError
    return bluetooth


@pytest.fixture
def mbluetooth_import(mbluetooth):
    orig_import = __import__

    def mocked_import(name, *args):
        if name == "bluetooth":
            return mbluetooth
        return orig_import(name, *args)

    with patch("builtins.__import__", new=mocked_import):
        yield mocked_import


@pytest.fixture
def mbluetooth_import_error():
    orig_import = __import__

    def mocked_import(name, *args):
        if name == "bluetooth":
            raise ImportError("mocked")
        return orig_import(name, *args)

    with patch("builtins.__import__", new=mocked_import):
        yield mocked_import


@pytest.fixture
def mbluetooth_import_not_supported():
    orig_import = __import__

    def mocked_import(name, *args):
        if name == "bluetooth":
            raise Exception("mocked")
        return orig_import(name, *args)

    with patch("builtins.__import__", new=mocked_import):
        yield mocked_import


def test_bluetooth(mbluetooth, mbluetooth_import, msock):
    # Instantiate backend.
    backend = nxt.backend.bluetooth.get_backend()
    # Find brick.
    mbluetooth.discover_devices.return_value = [
        "00:01:02:03:04:05",
    ]
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 1
    brick = bricks[0]
    assert mbluetooth.BluetoothSocket.called
    assert msock.connect.called
    sock = brick._sock
    # str.
    assert str(sock) == "Bluetooth (00:01:02:03:04:05)"
    # Send.
    some_bytes = bytes.fromhex("01020304")
    some_len = bytes.fromhex("0400")
    some_bytes_with_len = some_len + some_bytes
    sock.send(some_bytes)
    assert msock.send.call_args == call(some_bytes_with_len)
    # Recv.
    msock.recv.side_effect = [some_len, some_bytes]
    r = sock.recv()
    assert r == some_bytes
    assert msock.recv.called
    # Close.
    brick.close()
    assert msock.close.called
    # Duplicated close.
    sock.close()


def test_bluetooth_by_name(mbluetooth, mbluetooth_import, msock):
    # Instantiate backend.
    backend = nxt.backend.bluetooth.get_backend()
    # Find brick.
    mbluetooth.discover_devices.return_value = [
        ("00:01:02:03:04:05", "NXT"),
        ("00:01:02:03:04:06", "NXT"),
        ("00:01:02:03:04:07", "NXT2"),
    ]
    bricks = list(backend.find(name="NXT2", blah="blah"))
    assert len(bricks) == 1
    brick = bricks[0]
    assert mbluetooth.BluetoothSocket.called
    assert msock.connect.called
    sock = brick._sock
    # str.
    assert str(sock) == "Bluetooth (00:01:02:03:04:07)"
    # Close.
    brick.close()


def test_bluetooth_by_host(mbluetooth, mbluetooth_import, msock):
    # Instantiate backend.
    backend = nxt.backend.bluetooth.get_backend()
    # Find brick.
    bricks = list(backend.find(host="00:01:02:03:04:05", blah="blah"))
    assert len(bricks) == 1
    brick = bricks[0]
    assert not mbluetooth.discover_devices.called
    assert mbluetooth.BluetoothSocket.called
    assert msock.connect.called
    sock = brick._sock
    # str.
    assert str(sock) == "Bluetooth (00:01:02:03:04:05)"
    # Close.
    brick.close()


def test_bluetooth_by_host_and_name(mbluetooth, mbluetooth_import, msock):
    # Instantiate backend.
    backend = nxt.backend.bluetooth.get_backend()
    # Find brick.
    bricks = list(backend.find(host="00:01:02:03:04:05", name="NXT"))
    assert len(bricks) == 1
    brick = bricks[0]
    assert not mbluetooth.discover_devices.called
    assert mbluetooth.BluetoothSocket.called
    assert msock.connect.called
    sock = brick._sock
    # str.
    assert str(sock) == "Bluetooth (00:01:02:03:04:05)"
    # Close.
    brick.close()


def test_bluetooth_not_present(mbluetooth_import_error):
    assert nxt.backend.bluetooth.get_backend() is None


def test_bluetooth_not_supported(mbluetooth_import_not_supported):
    assert nxt.backend.bluetooth.get_backend() is None


def test_bluetooth_cant_connect(mbluetooth, mbluetooth_import, msock):
    backend = nxt.backend.bluetooth.get_backend()
    mbluetooth.discover_devices.return_value = [
        "00:01:02:03:04:05",
    ]
    msock.connect.side_effect = [mbluetooth.BluetoothError]
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 0


def test_bluetooth_cant_discover(mbluetooth, mbluetooth_import, msock):
    backend = nxt.backend.bluetooth.get_backend()
    mbluetooth.discover_devices.side_effect = [OSError]
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 0
    mbluetooth.discover_devices.side_effect = [mbluetooth.BluetoothError]
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 0


@pytest.mark.nxt("bluetooth")
def test_bluetooth_real():
    # Instantiate backend.
    backend = nxt.backend.bluetooth.get_backend()
    # Find brick.
    bricks = list(backend.find())
    assert len(bricks) > 0, "no NXT found"
    brick = bricks[0]
    sock = brick._sock
    # str.
    assert str(sock).startswith("Bluetooth (")
    # Send.
    sock.send(bytes.fromhex("019b"))
    # Recv.
    r = sock.recv()
    assert r.startswith(bytes.fromhex("029b00"))
    # Close.
    brick.close()



#!/usr/bin/python3
"""NXT-Python tutorial: turn a motor."""
import nxt.locator
import nxt.motor

with nxt.locator.find() as b:
    # Get the motor connected to the port A.
    mymotor = b.get_motor(nxt.motor.Port.A)
    # Full circle in one direction.
    mymotor.turn(25, 360)
    # Full circle in the opposite direction.
    mymotor.turn(-25, 360)



# nxt.sensor.generic module -- Support for LEGO Mindstorms NXT sensors
# Copyright (C) 2006,2007  Douglas P Lau
# Copyright (C) 2009  Marcus Wanner, Paulo Vieira, rhn
# Copyright (C) 2010  melducky, Marcus Wanner
# Copyright (C) 2022  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""The :mod:`nxt.sensor.generic` module supports the sensors manufactured by LEGO."""
import enum

from nxt.sensor import Mode, Type
from nxt.sensor.analog import BaseAnalogSensor
from nxt.sensor.digital import BaseDigitalSensor


class Touch(BaseAnalogSensor):
    """LEGO Mindstorms NXT touch sensor, part number 53793."""

    def __init__(self, brick, port):
        super().__init__(brick, port)
        self.set_input_mode(Type.SWITCH, Mode.BOOL)

    def is_pressed(self):
        """Get sensor pressed state.

        :return: ``True`` if the sensor is pressed, else ``False``.
        :rtype: bool
        """
        return bool(self.get_input_values().scaled_value)

    get_sample = is_pressed


class Light(BaseAnalogSensor):
    """LEGO Mindstorms NXT light sensor, part number 55969.

    This sensor is included in the first version of the Mindstorms NXT set and in the
    Education NXT set. Not to be confused with the color sensor included in the NXT 2.0
    set.

    :param bool illuminated: Initial illuminated state.
    """

    def __init__(self, brick, port, illuminated=True):
        super().__init__(brick, port)
        self.set_illuminated(illuminated)

    def set_illuminated(self, active):
        """Set illuminated state.

        :param bool active: ``True`` to activate light.
        """
        if active:
            type_ = Type.LIGHT_ACTIVE
        else:
            type_ = Type.LIGHT_INACTIVE
        self.set_input_mode(type_, Mode.RAW)

    def get_lightness(self):
        """Get detected light level.

        :return: Detected light level between 0 and 1023.
        :rtype: int
        """
        return self.get_input_values().scaled_value

    get_sample = get_lightness


class Sound(BaseAnalogSensor):
    """LEGO Mindstorms NXT sound sensor, part number 55963.

    :param bool adjusted: Initial adjusted state.
    """

    def __init__(self, brick, port, adjusted=True):
        super().__init__(brick, port)
        self.set_adjusted(adjusted)

    def set_adjusted(self, active):
        """Set adjusted state. When activated, the sensitivity of the sensor is adjusted
        to human perception.

        :param bool active: ``True`` to activate adjustment.
        """
        if active:
            type_ = Type.SOUND_DBA
        else:
            type_ = Type.SOUND_DB
        self.set_input_mode(type_, Mode.RAW)

    def get_loudness(self):
        """Get detected sound level.

        :return: Detected sound level between 0 and 1023.
        :rtype: int
        """
        return self.get_input_values().scaled_value

    get_sample = get_loudness


class Ultrasonic(BaseDigitalSensor):
    """LEGO Mindstorms NXT ultrasonic distance sensor, part number 53792."""

    I2C_ADDRESS = BaseDigitalSensor.I2C_ADDRESS.copy()
    I2C_ADDRESS.update(
        {
            "measurement_units": (0x14, "7s"),
            "continuous_measurement_interval": (0x40, "B"),
            "command": (0x41, "B"),
            "measurement_byte_0": (0x42, "B"),
            "measurement_byte_1": (0x43, "B"),
            "measurement_byte_2": (0x44, "B"),
            "measurement_byte_3": (0x45, "B"),
            "measurement_byte_4": (0x46, "B"),
            "measurement_byte_5": (0x47, "B"),
            "measurement_byte_6": (0x48, "B"),
            "measurement_byte_7": (0x49, "B"),
            "measurements": (0x42, "8B"),
            "actual_scale_factor": (0x51, "B"),
            "actual_scale_divisor": (0x52, "B"),
        }
    )

    class Command(enum.Enum):
        """Sensor command."""

        OFF = 0x00
        """Turn sensor OFF."""

        SINGLE_SHOT = 0x01
        """Enable single shot mode."""

        CONTINUOUS_MEASUREMENT = 0x02
        """Enable continuous measurement mode (default)."""

        EVENT_CAPTURE = 0x03
        """Enable event capture mode, to measure whether any other sensor is in the
        neighbourhood."""

        REQUEST_WARM_RESET = 0x04
        """Reset sensor."""

    def __init__(self, brick, port, check_compatible=True):
        super().__init__(brick, port, check_compatible)

    def get_distance(self):
        """Get distance to detected object.

        :return: Distance in cm.
        :rtype: int
        """
        return self.read_value("measurement_byte_0")[0]

    get_sample = get_distance

    def get_measurement_units(self):
        """Get measurement units.

        :return: String representing the measurement units, should be "10E-2m".
        :rtype: str
        """
        return (
            self.read_value("measurement_units")[0]
            .decode("windows-1252")
            .split("\0")[0]
        )

    def get_all_measurements(self):
        """Get all the measurements.

        :return: Eight distances in cm.
        :rtype: (int, int, int, int, int, int, int, int)
        """
        return self.read_value("measurements")

    def get_measurement_no(self, number):
        """Get one of the measurements.

        :param int number: Measurement number.
        :return: Distance in cm.
        :rtype: int
        """
        return self.read_value(f"measurement_byte_{number}")[0]

    def command(self, command):
        """Send a command to the sensor.

        :param Command command: Command to send.
        """
        self.write_value("command", (command.value,))

    def get_interval(self):
        """Get measurement interval, unknown unit.

        :return: Content of interval measurement register.
        :rtype: int
        """
        return self.read_value("continuous_measurement_interval")[0]

    def set_interval(self, interval):
        """Set measurement interval, unknown unit.

        :param int interval: New content of interval measurement register.
        """
        self.write_value("continuous_measurement_interval", (interval,))


Ultrasonic.add_compatible_sensor(None, "LEGO", "Sonar")  # Tested with version 'V1.0'


class Color(BaseAnalogSensor):
    """LEGO Mindstorms NXT color sensor, part number 64892.

    This sensor is included in the Mindstorms NXT 2.0 set. Not to be confused with the
    light sensor included in the first version or in the Education set.
    """

    class DetectedColor(enum.IntEnum):
        """Color detected by the sensor.

        This is an :class:`enum.IntEnum` for backward compatibility.
        """

        BLACK = 1
        """Black or low intensity."""

        BLUE = 2
        """Blue."""

        GREEN = 3
        """Green."""

        YELLOW = 4
        """Yellow."""

        RED = 5
        """Red."""

        WHITE = 6
        """White."""

    def __init__(self, brick, port):
        super().__init__(brick, port)
        self.set_light_color(Type.COLOR_FULL)

    def set_light_color(self, color):
        """Set the light color.

        :param Type color: Light color, or off, must be one of the Type.COLOR_* values.
        """
        self.set_input_mode(color, Mode.RAW)

    def get_light_color(self):
        """Get the light color previously set.

        :return: Light color, one of the Type.COLOR_* values.
        :rtype: Type
        """
        return self.get_input_values().sensor_type

    def get_reflected_light(self, color):
        """Get detected light level.

        :param Type color: Light color, or off, must be one of the Type.COLOR_* values.
        :return: Detected light level between 0 and 1023.
        :rtype: int
        """
        self.set_light_color(color)
        return self.get_input_values().scaled_value

    def get_color(self):
        """Get detected color.

        :return: Detected color.
        :rtype: DetectedColor

        This also sets the sensor mode to color detection (light is cycling quickly to
        determine color).
        """
        self.get_reflected_light(Type.COLOR_FULL)
        return self.DetectedColor(self.get_input_values().scaled_value)

    get_sample = get_color


class Temperature(BaseDigitalSensor):
    """LEGO Mindstorms NXT temperature sensor, part number 62840."""

    # This is actually a TI TMP275 chip: http://www.ti.com/product/tmp275
    I2C_DEV = 0x98
    I2C_ADDRESS = {"raw_value": (0x00, ">h")}

    def __init__(self, brick, port):
        # This sensor does not follow the convention of having version/vendor/product at
        # I2C registers 0x00/0x08/0x10, so check_compatible is always False.
        super().__init__(brick, port, False)

    def _get_raw_value(self):
        """Get raw unscaled value.

        :return: Unscaled value.
        :rtype: int
        """
        # This is a 12-bit value.
        return self.read_value("raw_value")[0] >> 4

    def get_deg_c(self):
        """Get the temperature in degrees Celsius.

        :return: Temperature in degrees Celsius.
        :rtype: float
        """
        v = self._get_raw_value()
        return round(v / 16, 1)

    def get_deg_f(self):
        """Get the temperature in degrees Fahrenheit.

        :return: Temperature in degrees Fahrenheit.
        :rtype: float
        """
        v = self._get_raw_value()
        return round(9 / 5 * v / 16 + 32, 1)

    get_sample = get_deg_c



# test_motor -- Test nxt.motor module
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
from unittest.mock import call

import pytest

import nxt.motor
from nxt.motor import Mode, Port, RegulationMode, RunState


@pytest.fixture
def mmotor_factory(mbrick):
    def factory(port):
        mbrick.get_output_state.return_value = (
            port,
            0,
            Mode.IDLE,
            RegulationMode.IDLE,
            0,
            RunState.IDLE,
            0,
            0,
            0,
            0,
        )
        m = mbrick.get_motor(port)
        assert mbrick.mock_calls == [call.get_output_state(port)]
        mbrick.reset_mock()
        return m

    return factory


@pytest.fixture
def mmotor(mmotor_factory):
    return mmotor_factory(Port.A)


@pytest.fixture
def mmotorb(mmotor_factory):
    return mmotor_factory(Port.B)


@pytest.fixture
def msyncmotor(mmotor, mmotorb):
    m = nxt.motor.SynchronizedMotors(mmotor, mmotorb, 50)
    return m


def test_reset_position(mbrick, mmotor):
    mmotor.reset_position(True)
    mmotor.reset_position(False)
    assert mbrick.mock_calls == [
        call.reset_motor_position(Port.A, True),
        call.reset_motor_position(Port.A, False),
    ]


def test_run(mbrick, mmotor):
    mmotor.run()
    assert mbrick.mock_calls == [
        # TODO: should be RegulationMode.IDLE.
        call.set_output_state(
            Port.A, 100, Mode.ON, RegulationMode.SPEED, 0, RunState.RUNNING, 0
        ),
    ]


def test_run_regulated(mbrick, mmotor):
    mmotor.run(50, regulated=True)
    assert mbrick.mock_calls == [
        call.set_output_state(
            Port.A,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SPEED,
            0,
            RunState.RUNNING,
            0,
        ),
    ]


def test_brake(mbrick, mmotor):
    mmotor.brake()
    assert mbrick.mock_calls == [
        call.set_output_state(
            Port.A,
            0,
            Mode.ON | Mode.BRAKE | Mode.REGULATED,
            RegulationMode.SPEED,
            0,
            RunState.RUNNING,
            0,
        ),
    ]


def test_idle(mbrick, mmotor):
    mmotor.idle()
    assert mbrick.mock_calls == [
        call.set_output_state(
            Port.A, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0
        ),
    ]


def test_weak_turn(mbrick, mmotor):
    mmotor.weak_turn(50, 360)
    assert mbrick.mock_calls == [
        call.set_output_state(
            Port.A, 50, Mode.ON, RegulationMode.IDLE, 0, RunState.RUNNING, 360
        ),
    ]


def test_turn(mbrick, mmotor, mtime):
    mbrick.get_output_state.side_effect = [
        (Port.A, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0, 0, 0, 0),
        # Test overshoot.
        (
            Port.A,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SPEED,
            0,
            RunState.RUNNING,
            0,
            720,
            720,
            720,
        ),
    ]
    mmotor.turn(50, 360)
    assert mbrick.mock_calls == [
        call.get_output_state(Port.A),
        call.set_output_state(
            Port.A,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SPEED,
            0,
            RunState.RUNNING,
            0,
        ),
        call.get_output_state(Port.A),
        call.set_output_state(
            Port.A,
            0,
            Mode.ON | Mode.REGULATED | Mode.BRAKE,
            RegulationMode.SPEED,
            0,
            RunState.RUNNING,
            0,
        ),
    ]


def test_turn_blocked(mbrick, mmotor, mtime):
    mtime.time.side_effect = [0, 1]
    with pytest.raises(nxt.motor.BlockedException):
        mmotor.turn(50, 360, brake=False, timeout=0.1, emulate=False)
    assert mbrick.mock_calls == [
        call.get_output_state(Port.A),
        call.set_output_state(
            Port.A,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SPEED,
            0,
            RunState.RUNNING,
            360,
        ),
        call.get_output_state(Port.A),
        call.set_output_state(
            Port.A, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0
        ),
    ]


def test_sync_run(mbrick, msyncmotor):
    msyncmotor.run(50)
    assert mbrick.mock_calls == [
        call.reset_motor_position(Port.A, True),
        call.reset_motor_position(Port.B, True),
        call.set_output_state(
            Port.A,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SYNC,
            # TODO: why reversed? This does not seems right.
            -50,
            RunState.RUNNING,
            0,
        ),
        call.set_output_state(
            Port.B,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SYNC,
            -50,
            RunState.RUNNING,
            0,
        ),
    ]


def test_sync_brake(mbrick, msyncmotor):
    msyncmotor.brake()
    assert mbrick.mock_calls == [
        # TODO: this should be possible to make it simpler.
        call.set_output_state(
            Port.A, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0
        ),
        call.set_output_state(
            Port.B, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0
        ),
        call.reset_motor_position(Port.A, True),
        call.reset_motor_position(Port.B, True),
        call.set_output_state(
            Port.A,
            0,
            Mode.ON | Mode.BRAKE | Mode.REGULATED,
            RegulationMode.SYNC,
            -50,
            RunState.RUNNING,
            0,
        ),
        call.set_output_state(
            Port.B,
            0,
            Mode.ON | Mode.BRAKE | Mode.REGULATED,
            RegulationMode.SYNC,
            -50,
            RunState.RUNNING,
            0,
        ),
        call.set_output_state(
            Port.A, 0, Mode.IDLE, RegulationMode.IDLE, -50, RunState.IDLE, 0
        ),
        call.set_output_state(
            Port.B, 0, Mode.IDLE, RegulationMode.IDLE, -50, RunState.IDLE, 0
        ),
        call.set_output_state(
            Port.A,
            0,
            Mode.ON | Mode.BRAKE | Mode.REGULATED,
            RegulationMode.SPEED,
            -50,
            RunState.RUNNING,
            0,
        ),
        call.set_output_state(
            Port.B,
            0,
            Mode.ON | Mode.BRAKE | Mode.REGULATED,
            RegulationMode.SPEED,
            -50,
            RunState.RUNNING,
            0,
        ),
    ]


def test_sync_idle(mbrick, msyncmotor):
    msyncmotor.idle()
    assert mbrick.mock_calls == [
        call.set_output_state(
            Port.A, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0
        ),
        call.set_output_state(
            Port.B, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0
        ),
    ]


def test_sync_turn(mbrick, msyncmotor):
    mbrick.get_output_state.side_effect = [
        (Port.A, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0, 0, 0, 0),
        (Port.B, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0, 0, 0, 0),
        # Test overshoot.
        (
            Port.A,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SYNC,
            0,
            RunState.RUNNING,
            0,
            720,
            720,
            720,
        ),
        (
            Port.B,
            0,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SYNC,
            0,
            RunState.RUNNING,
            0,
            0,
            0,
            0,
        ),
    ]
    msyncmotor.turn(50, 360, brake=False)
    assert mbrick.mock_calls == [
        call.reset_motor_position(Port.A, True),
        call.reset_motor_position(Port.B, True),
        call.get_output_state(Port.A),
        call.get_output_state(Port.B),
        call.set_output_state(
            Port.A,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SYNC,
            -50,
            RunState.RUNNING,
            0,
        ),
        call.set_output_state(
            Port.B,
            50,
            Mode.ON | Mode.REGULATED,
            RegulationMode.SYNC,
            -50,
            RunState.RUNNING,
            0,
        ),
        call.get_output_state(Port.A),
        call.get_output_state(Port.B),
        call.set_output_state(
            Port.A, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0
        ),
        call.set_output_state(
            Port.B, 0, Mode.IDLE, RegulationMode.IDLE, 0, RunState.IDLE, 0
        ),
    ]



# test_backend_devfile -- Test nxt.backend.devfile module
# Copyright (C) 2021  Nicolas Schodet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
from unittest.mock import Mock, call, patch

import pytest

import nxt.backend.devfile


@pytest.fixture
def mdev():
    dev = Mock(
        spec_set=(
            "read",
            "write",
            "close",
        )
    )
    return dev


@pytest.fixture
def mopen(mdev):
    with patch("nxt.backend.devfile.open") as fopen:
        fopen.return_value = mdev
        yield fopen


@pytest.fixture
def mglob():
    with patch("nxt.backend.devfile.glob") as glob:
        yield glob


@pytest.fixture
def mtty():
    with patch("nxt.backend.devfile.tty") as tty:
        yield tty


@pytest.fixture
def mplatform():
    with patch("nxt.backend.devfile.platform") as platform:
        yield platform


def test_devfile(mopen, mtty, mdev):
    # Instantiate backend.
    backend = nxt.backend.devfile.get_backend()
    # Find brick.
    bricks = list(backend.find(filename="/dev/nxt", blah="blah"))
    assert len(bricks) == 1
    brick = bricks[0]
    assert mopen.called
    assert mtty.setraw.called
    sock = brick._sock
    # str.
    assert str(sock).startswith("DevFile (/dev")
    # Send.
    some_bytes = bytes.fromhex("01020304")
    some_len = bytes.fromhex("0400")
    some_bytes_with_len = some_len + some_bytes
    sock.send(some_bytes)
    assert mdev.write.call_args == call(some_bytes_with_len)
    # Recv.
    mdev.read.side_effect = [some_len, some_bytes]
    r = sock.recv()
    assert r == some_bytes
    assert mdev.read.called
    # Close.
    brick.close()
    assert mdev.close.called
    # Duplicated close.
    sock.close()


def test_devfile_linux(mopen, mtty, mglob, mplatform):
    mplatform.system.return_value = "Linux"
    mglob.glob.return_value = ["/dev/rfcomm0"]
    backend = nxt.backend.devfile.get_backend()
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 1
    assert mglob.mock_calls == [call.glob("/dev/rfcomm*")]


def test_devfile_darwin(mopen, mtty, mglob, mplatform):
    mplatform.system.return_value = "Darwin"
    mglob.glob.return_value = ["/dev/tty.NXT-DevB-1"]
    backend = nxt.backend.devfile.get_backend()
    bricks = list(backend.find(name="NXT", blah="blah"))
    assert len(bricks) == 1
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 1
    assert mglob.mock_calls == [
        call.glob("/dev/*NXT*"),
        call.glob("/dev/*-DevB*"),
    ]


def test_devfile_other(mplatform):
    mplatform.system.return_value = "MSDos"
    backend = nxt.backend.devfile.get_backend()
    bricks = list(backend.find(blah="blah"))
    assert len(bricks) == 0


def test_devfile_cant_connect(mopen, mtty, mdev):
    backend = nxt.backend.devfile.get_backend()
    mopen.side_effect = [OSError]
    bricks = list(backend.find(filename="/dev/nxt", blah="blah"))
    assert len(bricks) == 0


@pytest.mark.nxt("devfile")
def test_devfile_real():
    # Instantiate backend.
    backend = nxt.backend.devfile.get_backend()
    # Find brick.
    bricks = list(backend.find())
    assert len(bricks) > 0, "no NXT found"
    brick = bricks[0]
    sock = brick._sock
    # str.
    assert str(sock).startswith("DevFile (/dev")
    # Send.
    sock.send(bytes.fromhex("019b"))
    # Recv.
    r = sock.recv()
    assert r.startswith(bytes.fromhex("029b00"))
    # Close.
    brick.close()



#!/usr/bin/python3
"""NXT-Python tutorial: find the brick."""
import nxt.locator

# Find a brick.
with nxt.locator.find() as b:
    # Once found, print its name.
    print("Found brick:", b.get_device_info()[0])
    # And play a recognizable note.
    b.play_tone(440, 250)
